from itertools import (
    groupby,
)
import json
from operator import (
    itemgetter,
)

from eth_abi import (
    encode,
    is_encodable,
    is_encodable_type,
)
from eth_abi.grammar import (
    parse,
)
from eth_utils import (
    keccak,
    to_tuple,
)

from .validation import (
    validate_structured_data,
)


def get_dependencies(primary_type, types):
    """
    Perform DFS to get all the dependencies of the primary_type.
    """
    deps = set()
    struct_names_yet_to_be_expanded = [primary_type]

    while len(struct_names_yet_to_be_expanded) > 0:
        struct_name = struct_names_yet_to_be_expanded.pop()

        deps.add(struct_name)
        fields = types[struct_name]
        for field in fields:
            field_type = field["type"]

            # Handle array types
            if is_array_type(field_type):
                field_type = field_type[: field_type.index("[")]

            if field_type not in types:
                # We don't need to expand types that are not user defined (customized)
                continue
            elif field_type not in deps:
                # Custom Struct Type
                struct_names_yet_to_be_expanded.append(field_type)
            elif field_type in deps:
                # skip types that we have already encountered
                continue
            else:
                raise TypeError(
                    f"Unable to determine type dependencies with type `{field_type}`."
                )
    # Don't need to make a struct as dependency of itself
    deps.remove(primary_type)

    return tuple(deps)


def field_identifier(field):
    """
    Convert a field dict into a typed-name string.

    Given a ``field`` of the format {'name': NAME, 'type': TYPE},
    this function converts it to ``TYPE NAME``
    """
    return f"{field['type']} {field['name']}"


def encode_struct(struct_name, struct_field_types):
    return f"{struct_name}({','.join(map(field_identifier, struct_field_types))})"


def encode_type(primary_type, types):
    """
    Serialize types into an encoded string.

    The type of a struct is encoded as:
    name ‖ "(" ‖ member₁ ‖ "," ‖ member₂ ‖ "," ‖ … ‖ memberₙ ")"
    where each member is written as type ‖ " " ‖ name.
    """
    # Getting the dependencies and sorting them alphabetically as per EIP712
    deps = get_dependencies(primary_type, types)
    sorted_deps = (primary_type,) + tuple(sorted(deps))

    result = "".join(
        [encode_struct(struct_name, types[struct_name]) for struct_name in sorted_deps]
    )
    return result


def hash_struct_type(primary_type, types):
    return keccak(text=encode_type(primary_type, types))


def is_array_type(type):
    return type.endswith("]")


@to_tuple
def get_depths_and_dimensions(data, depth):
    """
    Yields 2-length tuples of depth and dimension of each element at that depth.
    """
    if not isinstance(data, (list, tuple)):
        # Not checking for Iterable instance, because even Dictionaries and strings
        # are considered as iterables, but that's not what we want the condition to be.
        return ()

    yield depth, len(data)

    for item in data:
        # iterating over all 1 dimension less sub-data items
        yield from get_depths_and_dimensions(item, depth + 1)


def get_array_dimensions(data):
    """
    Given an array type data item, check that it is an array and return the dimensions
    as a tuple, in order from inside to outside.

    Ex: get_array_dimensions([[1, 2, 3], [4, 5, 6]]) returns (3, 2)
    """
    depths_and_dimensions = get_depths_and_dimensions(data, 0)

    # re-form as a dictionary with `depth` as key, and all of the dimensions
    # found at that depth.
    grouped_by_depth = {
        depth: tuple(dimension for depth, dimension in group)
        for depth, group in groupby(depths_and_dimensions, itemgetter(0))
    }

    dimensions = tuple(
        # check that all dimensions are the same, else use "dynamic"
        dimensions[0] if all(dim == dimensions[0] for dim in dimensions) else "dynamic"
        for _depth, dimensions in sorted(grouped_by_depth.items(), reverse=True)
    )

    return dimensions


def encode_field(types, name, field_type, value):
    if value is None:
        raise ValueError(f"Missing value for field {name} of type {field_type}")

    if field_type in types:
        return ("bytes32", keccak(encode_data(field_type, types, value)))

    if field_type == "bytes":
        if not isinstance(value, bytes):
            raise TypeError(
                f"Value of field `{name}` ({value}) is of the type `{type(value)}`, "
                f"but expected bytes value"
            )

        return ("bytes32", keccak(value))

    if field_type == "string":
        if not isinstance(value, str):
            raise TypeError(
                f"Value of field `{name}` ({value}) is of the type `{type(value)}`, "
                f"but expected string value"
            )

        return ("bytes32", keccak(text=value))

    if is_array_type(field_type):
        # Get the dimensions from the value
        array_dimensions = get_array_dimensions(value)
        # Get the dimensions from what was declared in the schema
        parsed_field_type = parse(field_type)

        for i in range(len(array_dimensions)):
            if len(parsed_field_type.arrlist[i]) == 0:
                # Skip empty or dynamically declared dimensions
                continue
            if array_dimensions[i] != parsed_field_type.arrlist[i][0]:
                # Dimensions should match with declared schema
                raise TypeError(
                    f"Array data `{value}` has dimensions `{array_dimensions}`"
                    f" whereas the schema has dimensions "
                    f"`{tuple(map(lambda x: x[0] if x else 'dynamic', parsed_field_type.arrlist))}`"  # noqa: E501
                )

        field_type_of_inside_array = field_type[: field_type.rindex("[")]
        field_type_value_pairs = [
            encode_field(types, name, field_type_of_inside_array, item)
            for item in value
        ]

        # handle empty array
        if value:
            data_types, data_hashes = zip(*field_type_value_pairs)
        else:
            data_types, data_hashes = [], []

        return ("bytes32", keccak(encode(data_types, data_hashes)))

    # First checking to see if field_type is valid as per abi
    if not is_encodable_type(field_type):
        raise TypeError(f"Received Invalid type `{field_type}` in field `{name}`")

    # Next, see if the value is encodable as the specified field_type
    if is_encodable(field_type, value):
        # field_type is a valid type and the provided value is encodable as that type
        return (field_type, value)
    else:
        raise TypeError(
            f"Value of `{name}` ({value}) is not encodable as type `{field_type}`. "
            f"If the base type is correct, verify that the value does not "
            f"exceed the specified size for the type."
        )


def encode_data(primary_type, types, data):
    encoded_types = ["bytes32"]
    encoded_values = [hash_struct_type(primary_type, types)]

    for field in types[primary_type]:
        type, value = encode_field(
            types, field["name"], field["type"], data[field["name"]]
        )
        encoded_types.append(type)
        encoded_values.append(value)

    return encode(encoded_types, encoded_values)


def load_and_validate_structured_message(structured_json_string_data):
    structured_data = json.loads(structured_json_string_data)
    validate_structured_data(structured_data)

    return structured_data


def hash_domain(structured_data):
    return keccak(
        encode_data("EIP712Domain", structured_data["types"], structured_data["domain"])
    )


def hash_message(structured_data):
    return keccak(
        encode_data(
            structured_data["primaryType"],
            structured_data["types"],
            structured_data["message"],
        )
    )



import os

from eth_utils import (
    ValidationError,
    to_bytes,
)
from hexbytes import (
    HexBytes,
)
import pytest
from toolz import (
    merge,
)

from eth_account import (
    Account,
)
from eth_account._utils.typed_transactions import (
    BlobTransaction,
)

TEST_DATA_PATH = os.path.join(os.path.dirname(__file__), "_test_data")
SIGNED_TX_PATH = os.path.join(TEST_DATA_PATH, "signed_tx.txt")
BLOB_DATA_1_PATH = os.path.join(TEST_DATA_PATH, "blob_data_1.txt")
BLOB_DATA_1_SIGNED_PATH = os.path.join(TEST_DATA_PATH, "blob_data_1_signed.txt")

TEST_ACCT = Account.from_key(
    "0x4646464646464646464646464646464646464646464646464646464646464646"
)

ZERO_BLOB = f"0x{'00' * 32 * 4096}"
ZERO_BLOB_VERSIONED_HASH = (
    "0x010657f37554c781402a22917dee2f75def7ab966d7b770905398eba3c444014"
)
ZERO_BLOB_COMMITMENT_AND_PROOF_HASH = "0xc00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"  # noqa: E501
BLOB_TX_DICT = {
    "chainId": 1,
    "nonce": 1,
    "maxPriorityFeePerGas": 50,
    "maxFeePerGas": 1000,
    "gas": 100000,
    "to": "0x45Ae5777c9b35Eb16280e423b0d7c91C06C66B58",
    "value": 1,
    "data": "0x52fdfc072182654f",
    "maxFeePerBlobGas": 100,
}


def test_validation_when_blobs_are_present_with_incorrect_versioned_hashes_in_tx_dict():
    tx_dict_with_wrong_versioned_hashes = merge(
        BLOB_TX_DICT, {"blobVersionedHashes": [f"0x{'00' * 32}"]}
    )
    with pytest.raises(
        ValidationError,
        match=(
            "`blobVersionedHashes` value defined in transaction does not match "
            "versioned hashes computed from blobs."
        ),
    ):
        BlobTransaction.from_dict(
            tx_dict_with_wrong_versioned_hashes,
            blobs=[to_bytes(hexstr=ZERO_BLOB)],
        )


def test_blobs_commitments_proofs_and_hashes_from_blobs():
    # test that when correct blobVersionedHashes value is passed into tx dict,
    # validation does not raise
    correct_versioned_hashes = [ZERO_BLOB_VERSIONED_HASH]
    tx_dict_with_correct_versioned_hashes = merge(
        BLOB_TX_DICT, {"blobVersionedHashes": correct_versioned_hashes}
    )

    # test does not raise
    tx = BlobTransaction.from_dict(
        tx_dict_with_correct_versioned_hashes, blobs=[to_bytes(hexstr=ZERO_BLOB)]
    )
    assert (
        len(tx.blob_data.blobs)
        == len(tx.blob_data.versioned_hashes)
        == len(tx.blob_data.proofs)
        == 1
    )
    # assert calculated versioned hash is the same as the provided versioned hash
    assert tx.blob_data.versioned_hashes[0].as_hexstr() == ZERO_BLOB_VERSIONED_HASH

    assert (
        tx.blob_data.commitments[0].as_hexstr() == ZERO_BLOB_COMMITMENT_AND_PROOF_HASH
    )
    assert tx.blob_data.proofs[0].as_hexstr() == ZERO_BLOB_COMMITMENT_AND_PROOF_HASH


def test_sign_blob_transaction_with_zero_blob_and_compare_with_tx_from_bytes():
    # test case taken from web3j library:
    # https://github.com/web3j/web3j/blob/c0679c08dfc1e1b196d1086f47d2eedc72f2a8f7/crypto/src/test/resources/blob_data/blob_tx_signed.txt  # noqa: E501
    # https://github.com/web3j/web3j/blob/c0679c08dfc1e1b196d1086f47d2eedc72f2a8f7/crypto/src/test/java/org/web3j/crypto/TransactionEncoderTest.java#L123-L128  # noqa: E501
    with open(SIGNED_TX_PATH) as signed_tx_file:
        signed_tx_from_file = to_bytes(hexstr=signed_tx_file.read().strip("\n"))

    signed_tx = TEST_ACCT.sign_transaction(BLOB_TX_DICT, blobs=[ZERO_BLOB])
    assert signed_tx.raw_transaction == HexBytes(signed_tx_from_file)

    # test `from_bytes()` creates `blob_data`
    tx_from_bytes = BlobTransaction.from_bytes(HexBytes(signed_tx_from_file))
    assert tx_from_bytes.blob_data is not None
    assert (
        len(tx_from_bytes.blob_data.blobs)
        == len(tx_from_bytes.blob_data.versioned_hashes)
        == len(tx_from_bytes.blob_data.proofs)
        == 1
    )
    assert tx_from_bytes.blob_data.blobs[0].as_hexstr() == ZERO_BLOB
    assert (
        tx_from_bytes.blob_data.versioned_hashes[0].as_hexbytes()
        == (tx_from_bytes.dictionary["blobVersionedHashes"][0])
        == HexBytes(ZERO_BLOB_VERSIONED_HASH)
    )
    assert (
        tx_from_bytes.blob_data.commitments[0].as_hexstr()
        == ZERO_BLOB_COMMITMENT_AND_PROOF_HASH
    )
    assert (
        tx_from_bytes.blob_data.proofs[0].as_hexstr()
        == ZERO_BLOB_COMMITMENT_AND_PROOF_HASH
    )


def test_blob_transaction_calculation_with_nonzero_blob():
    # test case taken from web3j library:
    # https://github.com/web3j/web3j/blob/c0679c08dfc1e1b196d1086f47d2eedc72f2a8f7/crypto/src/test/resources/blob_data/blob_data_2.txt  # noqa: E501
    with open(BLOB_DATA_1_PATH) as blob_data_1_file:
        blob_data_1 = to_bytes(hexstr=blob_data_1_file.read().strip("\n"))

    tx = BlobTransaction.from_dict(BLOB_TX_DICT, blobs=[blob_data_1])
    assert (
        len(tx.blob_data.blobs)
        == len(tx.blob_data.versioned_hashes)
        == len(tx.blob_data.proofs)
        == 1
    )
    assert tx.blob_data.blobs[0].as_bytes() == blob_data_1
    assert tx.blob_data.versioned_hashes[0].as_hexstr() == (
        "0x018ef96865998238a5e1783b6cafbc1253235d636f15d318f1fb50ef6a5b8f6a"
    )
    assert tx.blob_data.proofs[0].as_hexstr() == (
        "0x963150f3ee4d5e5f065429f587b4fa199cd8a866b8f6388eb52372870052603c98194c6521077c3260c41bf3b796c833"  # noqa: E501
    )

    with open(BLOB_DATA_1_SIGNED_PATH) as blob_data_1_signed_file:
        blob_data_1_signed = to_bytes(hexstr=blob_data_1_signed_file.read().strip("\n"))

    signed = TEST_ACCT.sign_transaction(tx.dictionary, blobs=[blob_data_1])
    assert blob_data_1_signed == signed.raw_transaction


def test_high_and_low_blob_count_limit_validation():
    with pytest.raises(ValidationError, match="must contain at least 1 blob"):
        BlobTransaction.from_dict(BLOB_TX_DICT, blobs=[])

    # make sure up to 6 blobs can be added to a blob transaction
    BlobTransaction.from_dict(BLOB_TX_DICT, blobs=[to_bytes(hexstr=ZERO_BLOB)] * 6)

    # assert raises if more than 6 blobs
    with pytest.raises(ValidationError, match="cannot contain more than 6 blobs"):
        BlobTransaction.from_dict(BLOB_TX_DICT, blobs=[to_bytes(hexstr=ZERO_BLOB)] * 7)



from typing import (
    Any,
    Dict,
    List,
    Tuple,
    Union,
)

from eth_abi import (
    encode,
)
from eth_utils import (
    keccak,
    to_bytes,
    to_int,
)

from eth_account._utils.encode_typed_data.helpers import (
    EIP712_SOLIDITY_TYPES,
    is_0x_prefixed_hexstr,
    is_array_type,
    parse_core_array_type,
    parse_parent_array_type,
)


def get_primary_type(types: Dict[str, List[Dict[str, str]]]) -> str:
    custom_types = set(types.keys())
    custom_types_that_are_deps = set()

    for type_ in custom_types:
        type_fields = types[type_]
        for field in type_fields:
            parsed_type = parse_core_array_type(field["type"])
            if parsed_type in custom_types and parsed_type != type_:
                custom_types_that_are_deps.add(parsed_type)

    primary_type = list(custom_types.difference(custom_types_that_are_deps))
    if len(primary_type) == 1:
        return primary_type[0]
    else:
        raise ValueError("Unable to determine primary type")


def encode_field(
    types: Dict[str, List[Dict[str, str]]],
    name: str,
    type_: str,
    value: Any,
) -> Tuple[str, Union[int, bytes]]:
    if type_ in types.keys():
        # type is a custom type
        if value is None:
            return ("bytes32", b"\x00" * 32)
        else:
            return ("bytes32", keccak(encode_data(type_, types, value)))

    elif type_ in ["string", "bytes"] and value is None:
        return ("bytes32", b"")

    # None is allowed only for custom and dynamic types
    elif value is None:
        raise ValueError(f"Missing value for field `{name}` of type `{type_}`")

    elif is_array_type(type_):
        # handle array type with non-array value
        if not isinstance(value, list):
            raise ValueError(
                f"Invalid value for field `{name}` of type `{type_}`: "
                f"expected array, got `{value}` of type `{type(value)}`"
            )

        parsed_type = parse_parent_array_type(type_)
        type_value_pairs = [
            encode_field(types, name, parsed_type, item) for item in value
        ]
        if not type_value_pairs:
            # the keccak hash of `encode((), ())`
            return (
                "bytes32",
                b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';{\xfa\xd8\x04]\x85\xa4p",  # noqa: E501
            )

        data_types, data_hashes = zip(*type_value_pairs)
        return ("bytes32", keccak(encode(data_types, data_hashes)))

    elif type_ == "bool":
        return (type_, bool(value))

    # all bytes types allow hexstr and str values
    elif type_.startswith("bytes"):
        if not isinstance(value, bytes):
            if is_0x_prefixed_hexstr(value):
                value = to_bytes(hexstr=value)
            elif isinstance(value, str):
                value = to_bytes(text=value)
            else:
                if isinstance(value, int) and value < 0:
                    value = 0

                value = to_bytes(value)

        return (
            # keccak hash if dynamic `bytes` type
            ("bytes32", keccak(value))
            if type_ == "bytes"
            # if fixed bytesXX type, do not hash
            else (type_, value)
        )

    elif type_ == "string":
        if isinstance(value, int):
            value = to_bytes(value)
        else:
            value = to_bytes(text=value)
        return ("bytes32", keccak(value))

    # allow string values for int and uint types
    elif isinstance(value, str) and type_.startswith(("int", "uint")):
        if is_0x_prefixed_hexstr(value):
            return (type_, to_int(hexstr=value))
        else:
            return (type_, to_int(text=value))

    return (type_, value)


def find_type_dependencies(type_, types, results=None):
    if results is None:
        results = set()

    # a type must be a string
    if not isinstance(type_, str):
        raise ValueError(
            "Invalid find_type_dependencies input: expected string, got "
            f"`{type_}` of type `{type(type_)}`"
        )
    # get core type if it's an array type
    type_ = parse_core_array_type(type_)

    if (
        # don't look for dependencies of solidity types
        type_ in EIP712_SOLIDITY_TYPES
        # found a type that's already been added
        or type_ in results
    ):
        return results

    # found a type that isn't defined
    elif type_ not in types:
        raise ValueError(f"No definition of type `{type_}`")

    results.add(type_)

    for field in types[type_]:
        find_type_dependencies(field["type"], types, results)
    return results


def encode_type(type_: str, types: Dict[str, List[Dict[str, str]]]) -> str:
    result = ""
    unsorted_deps = find_type_dependencies(type_, types)
    if type_ in unsorted_deps:
        unsorted_deps.remove(type_)

    deps = [type_] + sorted(list(unsorted_deps))
    for type_ in deps:
        children_list = []
        for child in types[type_]:
            child_type = child["type"]
            child_name = child["name"]
            children_list.append(f"{child_type} {child_name}")

        result += f"{type_}({','.join(children_list)})"

    return result


def hash_type(type_: str, types: Dict[str, List[Dict[str, str]]]) -> bytes:
    return bytes(keccak(text=encode_type(type_, types)))


def encode_data(
    type_: str,
    types: Dict[str, List[Dict[str, str]]],
    data: Dict[str, Any],
) -> bytes:
    encoded_types: List[str] = ["bytes32"]
    encoded_values: List[Union[bytes, int]] = [hash_type(type_, types)]

    for field in types[type_]:
        type, value = encode_field(
            types, field["name"], field["type"], data.get(field["name"])
        )
        encoded_types.append(type)
        encoded_values.append(value)

    return bytes(encode(encoded_types, encoded_values))


def hash_struct(
    type_: str,
    types: Dict[str, List[Dict[str, str]]],
    data: Dict[str, Any],
) -> bytes:
    encoded = encode_data(type_, types, data)
    return bytes(keccak(encoded))


def hash_eip712_message(
    # returns the same hash as `hash_struct`, but automatically determines primary type
    message_types: Dict[str, List[Dict[str, str]]],
    message_data: Dict[str, Any],
) -> bytes:
    primary_type = get_primary_type(message_types)
    return bytes(keccak(encode_data(primary_type, message_types, message_data)))


def hash_domain(domain_data: Dict[str, Any]) -> bytes:
    eip712_domain_map = {
        "name": {"name": "name", "type": "string"},
        "version": {"name": "version", "type": "string"},
        "chainId": {"name": "chainId", "type": "uint256"},
        "verifyingContract": {"name": "verifyingContract", "type": "address"},
        "salt": {"name": "salt", "type": "bytes32"},
    }

    for k in domain_data.keys():
        if k not in eip712_domain_map.keys():
            raise ValueError(f"Invalid domain key: `{k}`")

    domain_types = {
        "EIP712Domain": [
            eip712_domain_map[k] for k in eip712_domain_map.keys() if k in domain_data
        ]
    }

    return hash_struct("EIP712Domain", domain_types, domain_data)



from copy import (
    deepcopy,
)
import json
import re
import time

from eth_utils import (
    ValidationError,
    keccak,
)
from hexbytes import (
    HexBytes,
)
import pytest

from eth_account import (
    Account,
)
from eth_account._utils.structured_data.hashing import (
    encode_data,
    encode_struct,
    encode_type,
    get_array_dimensions,
    get_dependencies,
    hash_domain,
    hash_message,
    hash_struct_type,
    load_and_validate_structured_message,
)
from eth_account._utils.structured_data.validation import (
    TYPE_REGEX,
)
from eth_account.messages import (
    _hash_eip191_message,
    encode_structured_data,
)


@pytest.fixture
def eip712_example_json_string():
    return open("tests/fixtures/valid_eip712_example.json").read()


@pytest.fixture
def eip712_example_types(eip712_example_json_string):
    return json.loads(eip712_example_json_string)["types"]


@pytest.fixture
def eip712_example_domain_type(eip712_example_json_string):
    return json.loads(eip712_example_json_string)["types"]["EIP712Domain"]


@pytest.fixture
def eip712_example_message(eip712_example_json_string):
    return json.loads(eip712_example_json_string)["message"]


@pytest.fixture
def eip712_example_domain(eip712_example_json_string):
    return json.loads(eip712_example_json_string)["domain"]


@pytest.fixture(params=("text", "dict", "primitive", "hexstr"))
def eip712_message_encodings(request, eip712_example_json_string):
    if request.param == "text":
        return {"text": eip712_example_json_string}
    elif request.param == "primitive":
        return {"primitive": eip712_example_json_string.encode()}
    elif request.param == "dict":
        return {"primitive": json.loads(eip712_example_json_string)}
    elif request.param == "hexstr":
        return {"hexstr": eip712_example_json_string.encode().hex()}
    else:
        raise Exception("Unreachable")


# --- EIP 712 example with added "cc" array field to Mail struct --- #


@pytest.fixture
def eip712_example_with_array_json_string():
    return open("tests/fixtures/valid_eip712_example_with_array.json").read()


@pytest.fixture
def eip712_example_with_array_types(eip712_example_with_array_json_string):
    return json.loads(eip712_example_with_array_json_string)["types"]


@pytest.fixture
def eip712_example_with_array_domain_type(eip712_example_with_array_json_string):
    return json.loads(eip712_example_with_array_json_string)["types"]["EIP712Domain"]


@pytest.fixture
def eip712_example_with_array_message(eip712_example_with_array_json_string):
    return json.loads(eip712_example_with_array_json_string)["message"]


@pytest.fixture
def eip712_example_with_array_domain(eip712_example_with_array_json_string):
    return json.loads(eip712_example_with_array_json_string)["domain"]


@pytest.fixture(params=("text", "dict", "primitive", "hexstr"))
def eip712_with_array_message_encodings(request, eip712_example_with_array_json_string):
    if request.param == "text":
        return {"text": eip712_example_with_array_json_string}
    elif request.param == "primitive":
        return {"primitive": eip712_example_with_array_json_string.encode()}
    elif request.param == "dict":
        return {"primitive": json.loads(eip712_example_with_array_json_string)}
    elif request.param == "hexstr":
        return {"hexstr": eip712_example_with_array_json_string.encode().hex()}
    else:
        raise Exception("Unreachable")


# --- EIP 712 example with added multiple array fields --- #


@pytest.fixture
def eip712_example_with_multi_array_json_string():
    return open("tests/fixtures/valid_eip712_example_with_multi_array.json").read()


@pytest.fixture
def eip712_example_with_multi_array_types(eip712_example_with_multi_array_json_string):
    return json.loads(eip712_example_with_multi_array_json_string)["types"]


@pytest.fixture
def eip712_example_with_multi_array_domain_type(
    eip712_example_with_multi_array_json_string,
):
    return json.loads(eip712_example_with_multi_array_json_string)["types"][
        "EIP712Domain"
    ]


@pytest.fixture
def eip712_example_with_multi_array_message(
    eip712_example_with_multi_array_json_string,
):
    return json.loads(eip712_example_with_multi_array_json_string)["message"]


@pytest.fixture
def eip712_example_with_multi_array_domain(eip712_example_with_multi_array_json_string):
    return json.loads(eip712_example_with_multi_array_json_string)["domain"]


@pytest.fixture(params=("text", "dict", "primitive", "hexstr"))
def eip712_with_multi_array_message_encodings(
    request, eip712_example_with_multi_array_json_string
):
    if request.param == "text":
        return {"text": eip712_example_with_multi_array_json_string}
    elif request.param == "primitive":
        return {"primitive": eip712_example_with_multi_array_json_string.encode()}
    elif request.param == "dict":
        return {"primitive": json.loads(eip712_example_with_multi_array_json_string)}
    elif request.param == "hexstr":
        return {"hexstr": eip712_example_with_multi_array_json_string.encode().hex()}
    else:
        raise Exception("Unreachable")


@pytest.mark.parametrize(
    "primary_type, types, eip712_data, expected_hex",
    (
        (
            "BoolInt",
            {
                "BoolInt": [
                    {"name": "bool", "type": "bool"},
                    {"name": "bool_a", "type": "bool[]"},
                    {"name": "int", "type": "uint256"},
                    {"name": "int_a", "type": "uint256[][3]"},
                ]
            },
            {
                "bool": True,
                "bool_a": [False, True],
                "int": 212,
                "int_a": [[212], [], [12, 24, 36]],
            },
            "8d06a59c4f3179fc688615a1d69d61ee86c9683a5c3af150874396b3021892200000000000"
            "000000000000000000000000000000000000000000000000000001a6eef7e35abe70267296"
            "41147f7915573c7e97b47efa546f5f6e3230263bcb49000000000000000000000000000000"
            "00000000000000000000000000000000d4eb11d95b7e83141a4db340518d5fc187abde6f73"
            "e97cb1a5c7854c23c8cdf8fa",
        ),
        (
            "StringBytes",
            {
                "StringBytes": [
                    {"name": "string", "type": "string"},
                    {"name": "string_a", "type": "string[]"},
                    {"name": "bytes", "type": "bytes"},
                    {"name": "bytes_a", "type": "bytes[]"},
                ]
            },
            {
                "string": "spam",
                "string_a": ["spam", "eggs"],
                "bytes": b"snekshak",
                "bytes_a": [b"shake", b"snake"],
            },
            "763a70467632fd4d114a365d4f03c5886ce9447950874f45cbd42421cbc5f17d"
            "000e3bc84207015e1ae7e42b8679963a82088323f7ef1b456c44eda274f579f6"
            "3b78c672c715dfcb45484f0c5a488c1a20679e7e4ec1d980f43165671b70a3e4"
            "9079dbeec4d62f4344389883b7fe5fcedc73483b9d6a4774ab0857bedd7dd14d"
            "7578ae1ed519e90d30b2e031348369679d8c09f47f717eb17de4bce1a7ad5d6b",
        ),
    ),
)
def test_encode_data_basic(primary_type, types, eip712_data, expected_hex):
    assert encode_data(primary_type, types, eip712_data).hex() == expected_hex


@pytest.mark.parametrize(
    "primary_type, types, eip712_data, error_type, error_message",
    (
        (
            "Missing",
            {
                "Missing": [
                    {"name": "bool", "type": "bool"},
                ]
            },
            {
                "bool": None,
            },
            ValueError,
            "Missing value for field bool of type bool",
        ),
        (
            "NotBytes",
            {
                "NotBytes": [
                    {"name": "bytes", "type": "bytes"},
                ]
            },
            {
                "bytes": 212,
            },
            TypeError,
            "Value of field `bytes` (212) is of the type `<class 'int'>`, "
            "but expected bytes value",
        ),
        (
            "NotString",
            {
                "NotString": [
                    {"name": "string", "type": "string"},
                ]
            },
            {
                "string": b"snakes",
            },
            TypeError,
            "Value of field `string` (b'snakes') is of the type `<class 'bytes'>`, "
            "but expected string value",
        ),
        (
            "NotEncodableType",
            {
                "NotEncodableType": [
                    {"name": "wat", "type": "wat"},
                ]
            },
            {
                "wat": "huh",
            },
            TypeError,
            "Received Invalid type `wat` in field `wat`",
        ),
        (
            "NotEncodableAsType",
            {
                "NotEncodableAsType": [
                    {"name": "wat", "type": "uint256"},
                ]
            },
            {
                "wat": True,
            },
            TypeError,
            "Value of `wat` (True) is not encodable as type `uint256`. "
            "If the base type is correct, verify that the value does not "
            "exceed the specified size for the type.",
        ),
    ),
)
def test_encode_data_error_messages(
    primary_type, types, eip712_data, error_type, error_message
):
    with pytest.raises(error_type) as e:
        assert encode_data(primary_type, types, eip712_data)
    assert str(e.value) == error_message


@pytest.mark.parametrize(
    "primary_type, expected",
    (
        ("Mail", ("Person",)),
        ("Person", ()),
    ),
)
def test_get_dependencies_eip712(primary_type, expected, eip712_example_types):
    assert get_dependencies(primary_type, eip712_example_types) == expected


@pytest.mark.parametrize(
    "primary_type, expected",
    (
        ("Mail", ("Person",)),
        ("Person", ()),
    ),
)
def test_get_dependencies_eip712_with_array(
    primary_type, expected, eip712_example_with_array_types
):
    assert get_dependencies(primary_type, eip712_example_with_array_types) == expected


@pytest.mark.parametrize(
    "primary_type, expected",
    (
        ("Mail", ("Person", "Company")),
        ("Person", ("Company",)),
        ("Company", ()),
    ),
)
def test_get_dependencies_eip712_with_multi_array(
    primary_type, expected, eip712_example_with_multi_array_types
):
    assert set(
        get_dependencies(primary_type, eip712_example_with_multi_array_types)
    ) == set(expected)


@pytest.mark.parametrize(
    "struct_name, expected",
    (
        ("Mail", "Mail(Person from,Person to,string contents)"),
        ("Person", "Person(string name,address wallet)"),
        (
            "EIP712Domain",
            "EIP712Domain(string name,string version,uint256 chainId,address verifyingContract)",  # noqa: E501
        ),
    ),
)
def test_encode_struct_eip712(struct_name, expected, eip712_example_types):
    assert encode_struct(struct_name, eip712_example_types[struct_name]) == expected


@pytest.mark.parametrize(
    "struct_name, expected",
    (
        ("Mail", "Mail(Person from,Person to,Person[] cc,string contents)"),
        ("Person", "Person(string name,address wallet)"),
        (
            "EIP712Domain",
            "EIP712Domain(string name,string version,uint256 chainId,address verifyingContract)",  # noqa: E501
        ),
    ),
)
def test_encode_struct_eip712_with_array(
    struct_name, expected, eip712_example_with_array_types
):
    assert (
        encode_struct(struct_name, eip712_example_with_array_types[struct_name])
        == expected
    )


@pytest.mark.parametrize(
    "struct_name, expected",
    (
        (
            "Mail",
            "Mail(uint256[] ids,Person from,Person to,Person[] cc,Person[][] bcc,string contents,string[3][4][2] tags)",  # noqa: E501
        ),
        (
            "Person",
            "Person(string name,Company company,string[] aliases,address wallet)",
        ),
        ("Company", "Company(string name,uint256 id)"),
        (
            "EIP712Domain",
            "EIP712Domain(string name,string version,uint256 chainId,address verifyingContract)",  # noqa: E501
        ),
    ),
)
def test_encode_struct_eip712_with_multi_array(
    struct_name, expected, eip712_example_with_multi_array_types
):
    assert (
        encode_struct(struct_name, eip712_example_with_multi_array_types[struct_name])
        == expected
    )


@pytest.mark.parametrize(
    "primary_type, expected",
    (
        (
            "Mail",
            "Mail(Person from,Person to,string contents)Person(string name,address wallet)",  # noqa: E501
        ),
        ("Person", "Person(string name,address wallet)"),
    ),
)
def test_encode_type_eip712(primary_type, expected, eip712_example_types):
    assert encode_type(primary_type, eip712_example_types) == expected


@pytest.mark.parametrize(
    "primary_type, expected",
    (
        (
            "Mail",
            "Mail(Person from,Person to,Person[] cc,string contents)Person(string name,address wallet)",  # noqa: E501
        ),
        ("Person", "Person(string name,address wallet)"),
    ),
)
def test_encode_type_eip712_with_array(
    primary_type, expected, eip712_example_with_array_types
):
    assert encode_type(primary_type, eip712_example_with_array_types) == expected


@pytest.mark.parametrize(
    "primary_type, expected",
    (
        (
            "Mail",
            "Mail(uint256[] ids,Person from,Person to,Person[] cc,Person[][] bcc,string contents,string[3][4][2] tags)Company(string name,uint256 id)Person(string name,Company company,string[] aliases,address wallet)",  # noqa: E501
        ),
        (
            "Person",
            "Person(string name,Company company,string[] aliases,address wallet)Company(string name,uint256 id)",  # noqa: E501
        ),
        ("Company", "Company(string name,uint256 id)"),
    ),
)
def test_encode_type_eip712_with_multi_array(
    primary_type, expected, eip712_example_with_multi_array_types
):
    assert encode_type(primary_type, eip712_example_with_multi_array_types) == expected


@pytest.mark.parametrize(
    "primary_type, expected_hex",
    (
        ("Mail", "a0cedeb2dc280ba39b857546d74f5549c3a1d7bdc2dd96bf881f76108e23dac2"),
        ("Person", "b9d8c78acf9b987311de6c7b45bb6a9c8e1bf361fa7fd3467a2163f994c79500"),
    ),
)
def test_hash_struct_type_eip712(primary_type, expected_hex, eip712_example_types):
    assert hash_struct_type(primary_type, eip712_example_types).hex() == expected_hex


@pytest.mark.parametrize(
    "primary_type, expected_hex",
    (
        ("Person", "b9d8c78acf9b987311de6c7b45bb6a9c8e1bf361fa7fd3467a2163f994c79500"),
        ("Mail", "b19c5c7781083ed7325660d847f8c547b744917d67b38858b00bce7e62a4866b"),
    ),
)
def test_hash_struct_type_eip712_with_array(
    primary_type, expected_hex, eip712_example_with_array_types
):
    assert (
        hash_struct_type(primary_type, eip712_example_with_array_types).hex()
        == expected_hex
    )


@pytest.mark.parametrize(
    "primary_type, expected_hex",
    (
        ("Company", "6ca77e7c32d0c89b799964b908f09a67d27441f2c23a6fb1fe12744a6ad629f7"),
        ("Person", "c7b964db9fb195cb39c83204fdf1f0f406b6a2f3e463db9cad34aa0c040380a9"),
        ("Mail", "1cf41a63c532d5c0c81650d839a38629cc2ab400a6a4bd000b4188adf9dd7540"),
    ),
)
def test_hash_struct_type_eip712_with_multi_array(
    primary_type, expected_hex, eip712_example_with_multi_array_types
):
    assert (
        hash_struct_type(primary_type, eip712_example_with_multi_array_types).hex()
        == expected_hex
    )


def test_encode_data_eip712(eip712_example_types, eip712_example_message):
    primary_type = "Mail"
    expected_hex = (
        "a0cedeb2dc280ba39b857546d74f5549c3a1d7bdc2dd96bf881f76108e23dac2fc71e5fa27ff"
        "56c350aa531bc129ebdf613b772b6604664f5d8dbe21b85eb0c8cd54f074a4af31b4411ff6a6"
        "0c9719dbd559c221c8ac3492d9d872b041d703d1b5aadf3154a261abdd9086fc627b61efca26"
        "ae5702701d05cd2305f7c52a2fc8"
    )
    assert (
        encode_data(primary_type, eip712_example_types, eip712_example_message).hex()
        == expected_hex
    )


def test_encode_data_eip712_with_array(
    eip712_example_with_array_types, eip712_example_with_array_message
):
    primary_type = "Mail"
    expected_hex = (
        "b19c5c7781083ed7325660d847f8c547b744917d67b38858b00bce7e62a4866bfc71e5fa27ff"
        "56c350aa531bc129ebdf613b772b6604664f5d8dbe21b85eb0c8cd54f074a4af31b4411ff6a6"
        "0c9719dbd559c221c8ac3492d9d872b041d703d105a7f44bc01de4515b63df31009d8ce4fe8b"
        "06eb06a1ff4eec0b562d1c703535b5aadf3154a261abdd9086fc627b61efca26ae5702701d05"
        "cd2305f7c52a2fc8"
    )
    assert (
        encode_data(
            primary_type,
            eip712_example_with_array_types,
            eip712_example_with_array_message,
        ).hex()
        == expected_hex
    )


def test_encode_data_eip712_with_multi_array(
    eip712_example_with_multi_array_types, eip712_example_with_multi_array_message
):
    primary_type = "Mail"
    expected_hex = (
        "1cf41a63c532d5c0c81650d839a38629cc2ab400a6a4bd000b4188adf9dd754082ebb1c2e604ce"
        "0ff44d8591ae5cf16cbafd622284d40b37011b12dca2e2d9e4a30aec196a390f2dc9de7320f587"
        "4214e5a7929e08a7a81d9ef96666fdcbab35e1f45419ba2d57832f702e17b344e470f556cda74e"
        "91e703303656404b77fec1c6c072aa6915cd93401ec3be997eedadb9bb3f6300ab5c9e61808efa"
        "bf7da54687a9c71c4a727b1ca222ac101d1fc58980c48ad66baac15c14a68bbaaced314fb5aadf"
        "3154a261abdd9086fc627b61efca26ae5702701d05cd2305f7c52a2fc888e7f5caefc6a01fd300"
        "aaa708d99f8c04ed97879c2fc16950cb82957ff79099"
    )
    assert (
        encode_data(
            primary_type,
            eip712_example_with_multi_array_types,
            eip712_example_with_multi_array_message,
        ).hex()
        == expected_hex
    )


def test_hash_struct_main_message_eip712(eip712_example_json_string):
    structured_data = json.loads(eip712_example_json_string)
    expected_hex = "c52c0ee5d84264471806290a3f2c4cecfc5490626bf912d01f240d7a274b371e"
    assert hash_message(structured_data).hex() == expected_hex


def test_hash_struct_main_message_eip712_with_array(
    eip712_example_with_array_json_string,
):
    structured_data = json.loads(eip712_example_with_array_json_string)
    expected_hex = "04e58c373e4cd0695f0c2d6e2d8e379d6b51b987cf27d582dd94a03532d430d3"
    assert hash_message(structured_data).hex() == expected_hex


def test_hash_struct_main_message_eip712_with_multi_array(
    eip712_example_with_multi_array_json_string,
):  # noqa: E501
    structured_data = json.loads(eip712_example_with_multi_array_json_string)
    expected_hex = "6c5faac29b3f4552147e5cfef75b4ac09fdc7668ba1208c5f153f047a0da4819"
    assert hash_message(structured_data).hex() == expected_hex


def test_hash_struct_domain_eip712(eip712_example_json_string):
    structured_data = json.loads(eip712_example_json_string)
    expected_hex = "f2cee375fa42b42143804025fc449deafd50cc031ca257e0b194a650a912090f"
    assert hash_domain(structured_data).hex() == expected_hex


def test_hash_struct_domain_eip712_with_array(eip712_example_with_array_json_string):
    structured_data = json.loads(eip712_example_with_array_json_string)
    expected_hex = "f2cee375fa42b42143804025fc449deafd50cc031ca257e0b194a650a912090f"
    assert hash_domain(structured_data).hex() == expected_hex


def test_hash_struct_domain_eip712_with_multi_array(
    eip712_example_with_multi_array_json_string,
):
    structured_data = json.loads(eip712_example_with_multi_array_json_string)
    expected_hex = "f2cee375fa42b42143804025fc449deafd50cc031ca257e0b194a650a912090f"
    assert hash_domain(structured_data).hex() == expected_hex


def test_hashed_structured_data_eip712(eip712_message_encodings):
    structured_msg = encode_structured_data(**eip712_message_encodings)
    hashed_structured_msg = _hash_eip191_message(structured_msg)
    expected_hex = "be609aee343fb3c4b28e1df9e632fca64fcfaede20f02e86244efddf30957bd2"
    assert hashed_structured_msg.hex() == expected_hex


def test_hashed_structured_data_eip712_with_array(eip712_with_array_message_encodings):
    structured_msg = encode_structured_data(**eip712_with_array_message_encodings)
    hashed_structured_msg = _hash_eip191_message(structured_msg)
    expected_hex = "1780e7e042fa9ec126ccb68cd707d61580d00601b3eff8a5ec05116b46007fdb"
    assert hashed_structured_msg.hex() == expected_hex


def test_hashed_structured_data_eip712_with_multi_array(
    eip712_with_multi_array_message_encodings,
):
    structured_msg = encode_structured_data(**eip712_with_multi_array_message_encodings)
    hashed_structured_msg = _hash_eip191_message(structured_msg)
    expected_hex = "da275374c3a0790d389b6490d6fdf499f39c72105423f09d7b9be62d2fc08671"
    assert hashed_structured_msg.hex() == expected_hex


def test_signature_verification_eip712(eip712_message_encodings):
    account = Account.create()
    structured_msg = encode_structured_data(**eip712_message_encodings)
    signed = Account.sign_message(structured_msg, account.key)
    new_addr = Account.recover_message(structured_msg, signature=signed.signature)
    assert new_addr == account.address


def test_signature_verification_eip712_with_array(eip712_with_array_message_encodings):
    account = Account.create()
    structured_msg = encode_structured_data(**eip712_with_array_message_encodings)
    signed = Account.sign_message(structured_msg, account.key)
    new_addr = Account.recover_message(structured_msg, signature=signed.signature)
    assert new_addr == account.address


def test_signature_verification_eip712_with_multi_array(
    eip712_with_multi_array_message_encodings,
):
    account = Account.create()
    structured_msg = encode_structured_data(**eip712_with_multi_array_message_encodings)
    signed = Account.sign_message(structured_msg, account.key)
    new_addr = Account.recover_message(structured_msg, signature=signed.signature)
    assert new_addr == account.address


def test_signature_variables_eip712(eip712_message_encodings):
    # Check that the signature of typed message is the same as that
    # mentioned in the EIP. The link is as follows
    # https://github.com/ethereum/EIPs/blob/master/assets/eip-712/Example.js
    structured_msg = encode_structured_data(**eip712_message_encodings)

    private_key = keccak(text="cow")
    acc = Account.from_key(private_key)
    assert HexBytes(acc.address) == HexBytes(
        "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826"
    )

    sig = Account.sign_message(structured_msg, private_key)
    assert sig.v == 28
    assert (
        hex(sig.r)
        == "0x4355c47d63924e8a72e509b65029052eb6c299d53a04e167c5775fd466751c9d"
    )
    assert (
        hex(sig.s)
        == "0x7299936d304c153f6443dfa05f40ff007d72911b6f72307f996231605b91562"
    )


def test_signature_variables_eip712_with_array(eip712_with_array_message_encodings):
    structured_msg = encode_structured_data(**eip712_with_array_message_encodings)

    private_key = keccak(text="cow")
    acc = Account.from_key(private_key)
    assert HexBytes(acc.address) == HexBytes(
        "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826"
    )

    sig = Account.sign_message(structured_msg, private_key)
    assert sig.v == 27
    assert (
        hex(sig.r)
        == "0x315bd45341016e306aa54a7aaf002104b7a5b51903be6ca04cef9d49555319f1"
    )
    assert (
        hex(sig.s)
        == "0x46afff92b3504a3ab106dd96c414e38709515d884433816801ddde8d10136c8c"
    )


def test_signature_variables_eip712_with_multi_array(
    eip712_with_multi_array_message_encodings,
):
    structured_msg = encode_structured_data(**eip712_with_multi_array_message_encodings)

    private_key = keccak(text="cow")
    acc = Account.from_key(private_key)
    assert HexBytes(acc.address) == HexBytes(
        "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826"
    )

    sig = Account.sign_message(structured_msg, private_key)
    assert sig.v == 27
    assert (
        hex(sig.r)
        == "0x873be7f5ac9688ea3675b0c44916340883103053de9d1f9d0c1b6e9758a78a2b"
    )
    assert (
        hex(sig.s)
        == "0x6309769ff21b7a4f610316b249b004d4f9e9e9d6df2634d9bc0f492cda6b9c06"
    )


def test_hashed_structured_data_with_bytes(eip712_example_with_array_json_string):
    structured_data = json.loads(eip712_example_with_array_json_string)

    # change 'contents' field to type ``bytes`` and set bytes value
    structured_data["types"]["Mail"][3]["type"] = "bytes"
    structured_data["message"]["contents"] = keccak(b"")

    structured_msg = encode_structured_data(structured_data)
    hashed_structured_msg = _hash_eip191_message(structured_msg)
    expected_hash_value_hex = (
        "6de2148c1b98a586b42eaac9bbd72c4af5ee490fc1ec53fe483667c8fb38d10d"
    )
    assert hashed_structured_msg.hex() == expected_hash_value_hex


def test_hashed_structured_data_with_bytes32(eip712_example_with_array_json_string):
    structured_data = json.loads(eip712_example_with_array_json_string)

    # change 'contents' field to type ``bytes32`` and set bytes value
    structured_data["types"]["Mail"][3]["type"] = "bytes32"
    structured_data["message"]["contents"] = keccak(b"")

    structured_msg = encode_structured_data(structured_data)
    hashed_structured_msg = _hash_eip191_message(structured_msg)
    expected_hash_value_hex = (
        "c9f862d796deff552a90d7de1faff9a4b7f99726bf7518dc00073d23373d0530"
    )
    assert hashed_structured_msg.hex() == expected_hash_value_hex


def test_hashed_structured_data_with_nested_structs():
    nested_structs_valid_data_json_string = open(
        "tests/fixtures/valid_message_nested_structs.json"
    ).read()

    structured_data = json.loads(nested_structs_valid_data_json_string)
    structured_msg = encode_structured_data(structured_data)
    hashed_structured_msg = _hash_eip191_message(structured_msg)

    expected_hash_value_hex = (
        "0c9b68928afbff5af30e99ae29fcdf1cc36974f7ba4a6e4336cc98daa4f0c2af"
    )
    assert hashed_structured_msg.hex() == expected_hash_value_hex


@pytest.mark.parametrize(
    "type, valid",
    (
        ("unint bytes32", False),
        ("hello\\[]", False),
        ("byte[]uint", False),
        ("byte[7[]uint][]", False),
        ("Person[0]", False),
        ("bytes32", True),
        ("Foo[]", True),
        ("bytes1", True),
        ("bytes32[][][]", True),
        ("byte[9]", True),
        ("Person[1]", True),
    ),
)
def test_type_regex(type, valid):
    if valid:
        assert re.match(TYPE_REGEX, type) is not None
    else:
        assert re.match(TYPE_REGEX, type) is None


def test_type_regex_for_redos():
    start = time.time()
    # len 30 string is long enough to cause > 1 second delay if the regex is bad
    long = "1" * 30
    invalid_structured_data_string = f"""{{
        "types": {{
            "EIP712Domain": [
                {{"name": "aaaa", "type": "$[{long}0"}},
                {{"name": "version", "type": "string"}},
                {{"name": "chainId", "type": "uint256"}},
                {{"name": "verifyingContract", "type": "address"}}
            ]
        }}
    }}"""

    with pytest.raises(re.error, match="unterminated character set at position 15"):
        with pytest.raises(
            ValidationError, match=f"Invalid Type `$[{long}0` in `EIP712Domain`"
        ):
            load_and_validate_structured_message(invalid_structured_data_string)

    done = time.time() - start
    assert done < 1


def test_structured_data_invalid_identifier_filtered_by_regex():
    invalid_structured_data_string = open(
        "tests/fixtures/invalid_struct_identifier_message.json"
    ).read()
    with pytest.raises(ValidationError) as e:
        load_and_validate_structured_message(invalid_structured_data_string)
    assert str(e.value) == "Invalid Identifier `hello wallet` in `Person`"


def test_structured_data_invalid_type_filtered_by_regex():
    invalid_structured_data_string = open(
        "tests/fixtures/invalid_struct_type_message.json"
    ).read()
    with pytest.raises(ValidationError) as e:
        load_and_validate_structured_message(invalid_structured_data_string)
    assert str(e.value) == "Invalid Type `Hello Person` in `Mail`"


def test_invalid_structured_data_value_type_mismatch_in_type():
    # Given type is valid (string), but the value (int) is not of the mentioned type
    invalid_structured_data_string = open(
        "tests/fixtures/invalid_message_value_type_mismatch_type.json"
    ).read()
    invalid_structured_data = json.loads(invalid_structured_data_string)
    with pytest.raises(
        TypeError,
        match="Value of field `contents` \\(12345\\) is of the type `<class 'int'>`, "
        "but expected string value",
    ):
        hash_message(invalid_structured_data)


def test_invalid_structured_data_invalid_abi_type():
    # Given type/types are invalid
    invalid_structured_data_string = open(
        "tests/fixtures/invalid_message_invalid_abi_type.json"
    ).read()
    invalid_structured_data = json.loads(invalid_structured_data_string)
    with pytest.raises(
        TypeError, match="Received Invalid type `uint25689` in field `balance`"
    ):
        hash_message(invalid_structured_data)


def test_structured_data_invalid_identifier_filtered_by_abi_encodable_function():
    # Given valid abi type, but the value is not encodable as the specified type
    # (Error is found by the ``is_encodable`` ABI function)
    invalid_structured_data_string = open(
        "tests/fixtures/invalid_message_valid_abi_type_invalid_value.json"
    ).read()
    invalid_structured_data = json.loads(invalid_structured_data_string)
    with pytest.raises(
        TypeError,
        match="Value of `balance` \\(how do you do\\?\\) is not encodable as type "
        "`uint256`. If the base type is correct, verify that the "
        "value does not exceed the specified size for the type.",
    ):
        hash_message(invalid_structured_data)


@pytest.mark.parametrize(
    "data, expected",
    (
        ([[1, 2, 3], [4, 5, 6]], (3, 2)),
        ([[1, 2, 3]], (3, 1)),
        ([1, 2, 3], (3,)),
        ([[[1, 2], [3, 4], [5, 6]], [[7, 8], [9, 10], [11, 12]]], (2, 3, 2)),
        # test dynamic when varying array sizes exist within the same dimension
        ([[212], [], [12, 24, 36]], ("dynamic", 3)),
        ([[212], [], [12, [22, 24], [36]]], ("dynamic", "dynamic", 3)),
    ),
)
def test_get_array_dimensions(data, expected):
    assert get_array_dimensions(data) == expected


def test_unequal_array_lengths_between_schema_and_data():
    invalid_structured_data_string = open(
        "tests/fixtures/invalid_message_unequal_1d_array_lengths.json"
    ).read()
    invalid_structured_data = json.loads(invalid_structured_data_string)
    with pytest.raises(TypeError) as e:
        hash_message(invalid_structured_data)
    assert str(e.value) == (
        "Array data "
        "`[{'name': 'Bob', 'wallet': '0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB'}]` "
        "has dimensions `(1,)` whereas the schema has dimensions `(2,)`"
    ) or str(e.value) == (
        "Array data "
        "`[{'wallet': '0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB', 'name': 'Bob'}]` "
        "has dimensions `(1,)` whereas the schema has dimensions `(2,)`"
    )


def test_unequal_array_dimension_between_schema_and_data():
    invalid_structured_data_string = open(
        "tests/fixtures/invalid_message_unequal_array_dimensions.json"
    ).read()
    invalid_structured_data = json.loads(invalid_structured_data_string)
    with pytest.raises(TypeError) as e:
        hash_message(invalid_structured_data)
    assert str(e.value) == (
        "Array data "
        "`[{'name': 'Bob', 'wallet': '0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB'}]` "
        "has dimensions `(1,)` whereas the schema has dimensions `(2, 'dynamic', 4)`"
    ) or str(e.value) == (
        "Array data "
        "`[{'wallet': '0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB', 'name': 'Bob'}]` "
        "has dimensions `(1,)` whereas the schema has dimensions `(2, 'dynamic', 4)`"
    )


def test_encode_structured_data_ignores_additional_data_in_a_custom_type():
    message_with_additonal_data = {
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
            ],
            "Person": [
                {"name": "name", "type": "string"},
            ],
        },
        "primaryType": "Person",
        "domain": {
            "name": "Name",
        },
        "message": {
            "name": "Bob",
            "pet": {
                "animal": "cat",
                "age": 3,
            },
        },
    }

    message_without_additonal_data = deepcopy(message_with_additonal_data)
    message_without_additonal_data["message"].pop("pet")
    assert message_without_additonal_data["message"].get("pet") is None

    assert encode_structured_data(
        message_with_additonal_data
    ) == encode_structured_data(message_without_additonal_data)


def test_encode_typed_data_ignores_unused_types():
    message_with_unused_type = {
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
            ],
            "Person": [
                {"name": "name", "type": "string"},
            ],
            "OtherPerson": [
                {"name": "name", "type": "string"},
            ],
        },
        "primaryType": "Person",
        "domain": {
            "name": "Name",
        },
        "message": {
            "name": "Bob",
        },
    }

    message_without_unused_type = deepcopy(message_with_unused_type)
    message_without_unused_type["types"].pop("OtherPerson")
    assert message_without_unused_type["types"].get("OtherPerson") is None

    assert encode_structured_data(message_with_unused_type) == encode_structured_data(
        message_without_unused_type
    )


def test_encode_typed_data_errors_on_unused_field_in_EIP712Domain():
    message_with_extra_EIP712Domain_field = {
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "number", "type": "uint256"},
            ],
            "Person": [
                {"name": "name", "type": "string"},
            ],
            "OtherPerson": [
                {"name": "name", "type": "string"},
            ],
        },
        "primaryType": "Person",
        "domain": {
            "name": "Name",
        },
        "message": {
            "name": "Bob",
        },
    }

    with pytest.raises(expected_exception=KeyError, match="number"):
        encode_structured_data(message_with_extra_EIP712Domain_field)


def test_encode_typed_data_ignores_extra_info_in_domain():
    message_with_extra_info_in_domain = {
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
            ],
            "Person": [
                {"name": "name", "type": "string"},
            ],
            "OtherPerson": [
                {"name": "name", "type": "string"},
            ],
        },
        "primaryType": "Person",
        "domain": {
            "name": "Name",
            "number": 27,
        },
        "message": {
            "name": "Bob",
        },
    }

    message_without_extra_info_in_domain = deepcopy(message_with_extra_info_in_domain)
    message_without_extra_info_in_domain["domain"].pop("number")
    assert message_without_extra_info_in_domain["domain"].get("number") is None

    assert encode_structured_data(
        message_with_extra_info_in_domain
    ) == encode_structured_data(message_without_extra_info_in_domain)



from abc import (
    ABC,
    abstractmethod,
)
import hashlib
import os
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from ckzg import (
    blob_to_kzg_commitment,
    compute_blob_kzg_proof,
    load_trusted_setup,
)
from eth_typing import (
    HexStr,
)
from eth_utils import (
    ValidationError,
    is_bytes,
    is_string,
    to_bytes,
    to_int,
)
from eth_utils.curried import (
    apply_formatter_to_array,
    apply_formatters_to_dict,
    apply_one_of_formatters,
    hexstr_if_str,
)
from eth_utils.toolz import (
    identity,
    merge,
)
from hexbytes import (
    HexBytes,
)
from pydantic import (
    BaseModel,
    ConfigDict,
    computed_field,
    field_validator,
)

from eth_account._utils.validation import (
    LEGACY_TRANSACTION_FORMATTERS,
)

TYPED_TRANSACTION_FORMATTERS = merge(
    LEGACY_TRANSACTION_FORMATTERS,
    {
        "chainId": hexstr_if_str(to_int),
        "type": hexstr_if_str(to_int),
        "accessList": apply_formatter_to_array(
            apply_formatters_to_dict(
                {
                    "address": apply_one_of_formatters(
                        (
                            (is_string, hexstr_if_str(to_bytes)),
                            (is_bytes, identity),
                        )
                    ),
                    "storageKeys": apply_formatter_to_array(hexstr_if_str(to_int)),
                }
            ),
        ),
        "maxPriorityFeePerGas": hexstr_if_str(to_int),
        "maxFeePerGas": hexstr_if_str(to_int),
        "maxFeePerBlobGas": hexstr_if_str(to_int),
        "blobVersionedHashes": apply_formatter_to_array(hexstr_if_str(to_bytes)),
    },
)

# import TRUSTED_SETUP from ./kzg_trusted_setup.txt
TRUSTED_SETUP = os.path.join(
    os.path.dirname(__file__), "blob_transactions", "kzg_trusted_setup.txt"
)
VERSIONED_HASH_VERSION_KZG = b"\x01"


class _BlobDataElement(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    data: HexBytes

    def as_hexbytes(self) -> HexBytes:
        return self.data

    def as_bytes(self) -> bytes:
        return bytes(self.data)

    def as_hexstr(self) -> HexStr:
        return HexStr(f"0x{self.as_bytes().hex()}")


class Blob(_BlobDataElement):
    """
    Represents a Blob.
    """

    @field_validator("data")
    def validate_data(cls, v: Union[HexBytes, bytes]) -> Union[HexBytes, bytes]:
        if len(v) != 4096 * 32:
            raise ValidationError(
                "Invalid Blob size. Blob data must be comprised of 4096 32-byte "
                "field elements."
            )
        return v


class BlobKZGCommitment(_BlobDataElement):
    """
    Represents a Blob KZG Commitment.
    """

    @field_validator("data")
    def validate_commitment(cls, v: Union[HexBytes, bytes]) -> Union[HexBytes, bytes]:
        if len(v) != 48:
            raise ValidationError("Blob KZG Commitment must be 48 bytes long.")
        return v


class BlobProof(_BlobDataElement):
    """
    Represents a Blob Proof.
    """

    @field_validator("data")
    def validate_proof(cls, v: Union[HexBytes, bytes]) -> Union[HexBytes, bytes]:
        if len(v) != 48:
            raise ValidationError("Blob Proof must be 48 bytes long.")
        return v


class BlobVersionedHash(_BlobDataElement):
    """
    Represents a Blob Versioned Hash.
    """

    @field_validator("data")
    def validate_versioned_hash(
        cls, v: Union[HexBytes, bytes]
    ) -> Union[HexBytes, bytes]:
        if len(v) != 32:
            raise ValidationError("Blob Versioned Hash must be 32 bytes long.")
        if v[:1] != VERSIONED_HASH_VERSION_KZG:
            raise ValidationError(
                "Blob Versioned Hash must start with the KZG version byte."
            )
        return v


class BlobPooledTransactionData(BaseModel):
    """
    Represents the blob data for a type 3 `PooledTransaction` as defined by
    EIP-4844. This class takes blobs as bytes and computes the corresponding
    commitments, proofs, and versioned hashes.
    """

    _versioned_hash_version_kzg: bytes = VERSIONED_HASH_VERSION_KZG
    _versioned_hashes: Optional[List[BlobVersionedHash]] = None
    _commitments: Optional[List[BlobKZGCommitment]] = None
    _proofs: Optional[List[BlobProof]] = None

    blobs: List[Blob]

    def _kzg_to_versioned_hash(self, kzg_commitment: BlobKZGCommitment) -> bytes:
        return (
            self._versioned_hash_version_kzg
            + hashlib.sha256(kzg_commitment.data).digest()[1:]
        )

    @field_validator("blobs")
    def validate_blobs(cls, v: List[Blob]) -> List[Blob]:
        if len(v) == 0:
            raise ValidationError("Blob transactions must contain at least 1 blob.")
        elif len(v) > 6:
            raise ValidationError("Blob transactions cannot contain more than 6 blobs.")
        return v

    # type ignored bc mypy does not support decorated properties
    # https://github.com/python/mypy/issues/1362
    @computed_field  # type: ignore
    @property
    def versioned_hashes(self) -> List[BlobVersionedHash]:
        if self._versioned_hashes is None:
            self._versioned_hashes = [
                BlobVersionedHash(
                    data=HexBytes(self._kzg_to_versioned_hash(commitment))
                )
                for commitment in self.commitments
            ]
        return self._versioned_hashes

    # type ignored bc mypy does not support decorated properties
    # https://github.com/python/mypy/issues/1362
    @computed_field  # type: ignore
    @property
    def commitments(self) -> List[BlobKZGCommitment]:
        if self._commitments is None:
            self._commitments = [
                BlobKZGCommitment(
                    data=HexBytes(
                        blob_to_kzg_commitment(
                            blob.data, load_trusted_setup(TRUSTED_SETUP)
                        )
                    )
                )
                for blob in self.blobs
            ]
        return self._commitments

    # type ignored bc mypy does not support decorated properties
    # https://github.com/python/mypy/issues/1362
    @computed_field  # type: ignore
    @property
    def proofs(self) -> List[BlobProof]:
        if self._proofs is None:
            self._proofs = [
                BlobProof(
                    data=HexBytes(
                        compute_blob_kzg_proof(
                            blob.data,
                            commitment.data,
                            load_trusted_setup(TRUSTED_SETUP),
                        )
                    )
                )
                for blob, commitment in zip(self.blobs, self.commitments)
            ]
        return self._proofs


class _TypedTransactionImplementation(ABC):
    """
    Abstract class that every typed transaction must implement.
    Should not be imported or used by clients of the library.
    """

    blob_data: Optional[BlobPooledTransactionData] = None

    @abstractmethod
    def hash(self) -> bytes:
        pass

    @abstractmethod
    def payload(self) -> bytes:
        pass

    @abstractmethod
    def as_dict(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def vrs(self) -> Tuple[int, int, int]:
        pass



from eth_account.signers.base import (
    BaseAccount,
)


class LocalAccount(BaseAccount):
    r"""
    A collection of convenience methods to sign and encrypt, with an
    embedded private key.

    :var bytes key: the 32-byte private key data

    .. code-block:: python

        >>> my_local_account.address
        "0xF0109fC8DF283027b6285cc889F5aA624EaC1F55"
        >>> my_local_account.key
        b"\x01\x23..."

    You can also get the private key by casting the account to :class:`bytes`:

    .. code-block:: python

        >>> bytes(my_local_account)
        b"\\x01\\x23..."
    """

    def __init__(self, key, account):
        """
        Initialize a new account with the given private key.

        :param eth_keys.PrivateKey key: to prefill in private key execution
        :param ~eth_account.account.Account account: the key-unaware management API
        """
        self._publicapi = account

        self._address = key.public_key.to_checksum_address()

        key_raw = key.to_bytes()
        self._private_key = key_raw

        self._key_obj = key

    @property
    def address(self):
        return self._address

    @property
    def key(self):
        """
        Get the private key.
        """
        return self._private_key

    def encrypt(self, password, kdf=None, iterations=None):
        """
        Generate a string with the encrypted key.

        This uses the same structure as in
        :meth:`~eth_account.account.Account.encrypt`, but without a
        private key argument.
        """
        return self._publicapi.encrypt(
            self.key, password, kdf=kdf, iterations=iterations
        )

    def signHash(self, message_hash):
        return self._publicapi.signHash(
            message_hash,
            private_key=self.key,
        )

    def unsafe_sign_hash(self, message_hash):
        return self._publicapi.unsafe_sign_hash(
            message_hash,
            private_key=self.key,
        )

    def sign_message(self, signable_message):
        """
        Generate a string with the encrypted key.

        This uses the same structure as in
        :meth:`~eth_account.account.Account.sign_message`, but without a
        private key argument.
        """
        return self._publicapi.sign_message(signable_message, private_key=self.key)

    def sign_transaction(self, transaction_dict, blobs=None):
        return self._publicapi.sign_transaction(transaction_dict, self.key, blobs=blobs)

    def __bytes__(self):
        return self.key



from typing import (
    Any,
    Dict,
    List,
    Tuple,
    cast,
)

from eth_rlp import (
    HashableRLP,
)
from eth_utils import (
    keccak,
)
from eth_utils.curried import (
    apply_formatters_to_dict,
)
from eth_utils.toolz import (
    dissoc,
    merge,
    partial,
    pipe,
)
from hexbytes import (
    HexBytes,
)
import rlp
from rlp.sedes import (
    BigEndianInt,
    Binary,
    CountableList,
    List as ListSedesClass,
    big_endian_int,
    binary,
)

from ..transaction_utils import (
    transaction_rlp_to_rpc_structure,
    transaction_rpc_to_rlp_structure,
)
from ..validation import (
    LEGACY_TRANSACTION_VALID_VALUES,
    is_int_or_prefixed_hexstr,
    is_rpc_structured_access_list,
)
from .base import (
    TYPED_TRANSACTION_FORMATTERS,
    _TypedTransactionImplementation,
)

# Define typed transaction common sedes.
# [[{20 bytes}, [{32 bytes}...]]...], where ... means
# “zero or more of the thing to the left”.
access_list_sede_type = CountableList(
    ListSedesClass(
        [
            Binary.fixed_length(20, allow_empty=False),
            CountableList(BigEndianInt(32)),
        ]
    ),
)


class AccessListTransaction(_TypedTransactionImplementation):
    """
    Represents an access list transaction per EIP-2930.
    """

    # This is the first transaction to implement the EIP-2718 typed transaction.
    transaction_type = 1  # '0x01'

    unsigned_transaction_fields = (
        ("chainId", big_endian_int),
        ("nonce", big_endian_int),
        ("gasPrice", big_endian_int),
        ("gas", big_endian_int),
        ("to", Binary.fixed_length(20, allow_empty=True)),
        ("value", big_endian_int),
        ("data", binary),
        ("accessList", access_list_sede_type),
    )

    signature_fields = (
        ("v", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    )

    transaction_field_defaults = {
        "type": b"0x1",
        "chainId": 0,
        "to": b"",
        "value": 0,
        "data": b"",
        "accessList": [],
    }

    _unsigned_transaction_serializer = type(
        "_unsigned_transaction_serializer",
        (HashableRLP,),
        {
            "fields": unsigned_transaction_fields,
        },
    )

    _signed_transaction_serializer = type(
        "_signed_transaction_serializer",
        (HashableRLP,),
        {
            "fields": unsigned_transaction_fields + signature_fields,
        },
    )

    def __init__(self, dictionary: Dict[str, Any]):
        self.dictionary = dictionary

    @classmethod
    def assert_valid_fields(cls, dictionary: Dict[str, Any]) -> None:
        transaction_valid_values = merge(
            LEGACY_TRANSACTION_VALID_VALUES,
            {
                "type": is_int_or_prefixed_hexstr,
                "accessList": is_rpc_structured_access_list,
            },
        )

        if "v" in dictionary and dictionary["v"] == 0:
            # This is insane logic that is required because the way we evaluate
            # correct types is in the `if not all()` branch below, and 0 obviously
            # maps to the int(0), which maps to False... This was not an issue in
            # non-typed transaction because v=0, couldn't exist with the chain offset.
            dictionary["v"] = "0x0"
        valid_fields = apply_formatters_to_dict(
            transaction_valid_values,
            dictionary,
        )  # type: Dict[str, Any]
        if not all(valid_fields.values()):
            invalid = {
                key: dictionary[key] for key, valid in valid_fields.items() if not valid
            }
            raise TypeError(f"Transaction had invalid fields: {repr(invalid)}")

    @classmethod
    def from_dict(
        cls, dictionary: Dict[str, Any], blobs: List[bytes] = None
    ) -> "AccessListTransaction":
        """
        Builds an AccessListTransaction from a dictionary.
        Verifies that the dictionary is well formed.
        """
        if blobs is not None:
            raise ValueError("Blob data is not supported for `AccessListTransaction`.")

        # Validate fields.
        cls.assert_valid_fields(dictionary)
        sanitized_dictionary = pipe(
            dictionary,
            dict,
            partial(merge, cls.transaction_field_defaults),
            apply_formatters_to_dict(TYPED_TRANSACTION_FORMATTERS),
        )

        # We have verified the type, we can safely remove it from the dictionary,
        # given that it is not to be included within the RLP payload.
        transaction_type = sanitized_dictionary.pop("type")
        if transaction_type != cls.transaction_type:
            raise ValueError(
                f"expected transaction type {cls.transaction_type}, "
                f"got {transaction_type}"
            )
        return cls(
            dictionary=sanitized_dictionary,
        )

    @classmethod
    def from_bytes(cls, encoded_transaction: HexBytes) -> "AccessListTransaction":
        """Builds an AccesslistTransaction from a signed encoded transaction."""
        if not isinstance(encoded_transaction, HexBytes):
            raise TypeError(f"expected Hexbytes, got type: {type(encoded_transaction)}")
        if not (
            len(encoded_transaction) > 0
            and encoded_transaction[0] == cls.transaction_type
        ):
            raise ValueError("unexpected input")
        # Format is (0x01 || TransactionPayload)
        # We strip the prefix, and RLP unmarshal the payload into our
        # signed transaction serializer.
        transaction_payload = encoded_transaction[1:]
        rlp_serializer = cls._signed_transaction_serializer
        dictionary = rlp_serializer.from_bytes(  # type: ignore
            transaction_payload
        ).as_dict()
        rpc_structured_dict = transaction_rlp_to_rpc_structure(dictionary)
        rpc_structured_dict["type"] = cls.transaction_type
        return cls.from_dict(rpc_structured_dict)

    def as_dict(self) -> Dict[str, Any]:
        """Returns this transaction as a dictionary."""
        dictionary = self.dictionary.copy()
        dictionary["type"] = self.__class__.transaction_type
        return dictionary

    def hash(self) -> bytes:
        """
        Hashes this AccessListTransaction to prepare it for signing.
        As per the EIP-2930 specifications, the signature is a secp256k1 signature over
        keccak256(0x01 || rlp([chainId, nonce, gasPrice, gasLimit, to, value, data, accessList])).  # noqa E501
        Here, we compute the keccak256(...) hash.
        """
        # Remove signature fields.
        transaction_without_signature_fields = dissoc(self.dictionary, "v", "r", "s")
        # RPC-structured transaction to rlp-structured transaction
        rlp_structured_txn_without_sig_fields = transaction_rpc_to_rlp_structure(
            transaction_without_signature_fields
        )
        rlp_serializer = self.__class__._unsigned_transaction_serializer
        hash = pipe(
            rlp_serializer.from_dict(rlp_structured_txn_without_sig_fields),  # type: ignore  # noqa: E501
            lambda val: rlp.encode(val),  # rlp([...])
            lambda val: bytes([self.__class__.transaction_type])
            + val,  # (0x01 || rlp([...]))
            keccak,  # keccak256(0x01 || rlp([...]))
        )
        return cast(bytes, hash)

    def payload(self) -> bytes:
        """
        Returns this transaction's payload as bytes.

        Here, the TransactionPayload = rlp([chainId,
        nonce, gasPrice, gasLimit, to, value, data, accessList,
        signatureYParity, signatureR, signatureS])
        """
        if not all(k in self.dictionary for k in "vrs"):
            raise ValueError("attempting to encode an unsigned transaction")
        rlp_serializer = self.__class__._signed_transaction_serializer
        rlp_structured_dict = transaction_rpc_to_rlp_structure(self.dictionary)
        payload = rlp.encode(
            rlp_serializer.from_dict(rlp_structured_dict)  # type: ignore
        )
        return cast(bytes, payload)

    def vrs(self) -> Tuple[int, int, int]:
        """Returns (v, r, s) if they exist."""
        if not all(k in self.dictionary for k in "vrs"):
            raise ValueError("attempting to encode an unsigned transaction")
        return (self.dictionary["v"], self.dictionary["r"], self.dictionary["s"])



import json
import os


def load_eip712_json(file_name):
    with open(os.path.join("tests/eip712_messages", file_name)) as f:
        return json.load(f)


VALID_FOR_ALL = load_eip712_json("valid_for_all.json")
VALID_FOR_PY_AND_ETHERS = load_eip712_json("valid_py_and_ethers.json")
VALID_FOR_PY_AND_METAMASK = load_eip712_json("valid_py_and_metamask.json")
INVALID = load_eip712_json("invalid.json")
ONE_ARG_INVALID = load_eip712_json("one_arg_invalid.json")

ALL_VALID_EIP712_MESSAGES = {
    **VALID_FOR_ALL,
    **VALID_FOR_PY_AND_ETHERS,
    **VALID_FOR_PY_AND_METAMASK,
}


def convert_to_3_arg(message):
    domain_data = message["domain"]
    message_types = message["types"]
    message_types.pop("EIP712Domain", None)
    message_data = message["message"]
    return domain_data, message_types, message_data



import re

from eth_abi.exceptions import (
    EncodingTypeError,
    ValueOutOfBounds,
)
import pytest

from eth_account._utils.encode_typed_data.encoding_and_hashing import (
    encode_data,
    encode_field,
    encode_type,
    find_type_dependencies,
    get_primary_type,
    hash_domain,
    hash_eip712_message,
    hash_struct,
    hash_type,
)


@pytest.mark.parametrize(
    "types, expected",
    (
        (
            {
                "Person": [
                    {"name": "name", "type": "string"},
                ],
            },
            "Person",
        ),
        (
            {
                "Person": [
                    {"name": "name", "type": "string"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            "Mail",
        ),
        (
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friend", "type": "Person"},
                ],
            },
            "Person",
        ),
        (
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friends", "type": "Person[]"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                    {"name": "attachments", "type": "Attachment[]"},
                ],
                "Attachment": [
                    {"name": "from", "type": "string"},
                ],
            },
            "Mail",
        ),
    ),
    ids=[
        "primary_type with no dependencies",
        "primary_type with one dependency",
        "primary_type with recursive dependency",
        "primary_type with array dependency",
    ],
)
def test_get_primary_type_pass(types, expected):
    assert get_primary_type(types) == expected


@pytest.mark.parametrize(
    "types, expected",
    (
        (
            {
                "Person": [
                    {"name": "other_person", "type": "OtherPerson"},
                ],
                "OtherPerson": [
                    {"name": "Person", "type": "Person"},
                ],
            },
            {
                "expected_exception": ValueError,
                "match": "Unable to determine primary type",
            },
        ),
        (
            {
                "Person": [
                    {"name": "name", "type": "string"},
                ],
                "Mail": [
                    {"name": "type", "type": "string"},
                ],
            },
            {
                "expected_exception": ValueError,
                "match": "Unable to determine primary type",
            },
        ),
    ),
    ids=[
        "primary_type circular dependency",
        "two primary types that do not depend on each other",
    ],
)
def test_get_primary_type_fail(types, expected):
    with pytest.raises(**expected):
        get_primary_type(types)


@pytest.mark.parametrize(
    "name, type_, value, expected",
    (
        (
            "custom_type",
            "Sample",
            None,
            (
                "bytes32",
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",  # noqa: E501
            ),
        ),
        (
            "name",
            "string",
            None,
            (
                "bytes32",
                b"",
            ),
        ),
        (
            "some_bytes",
            "bytes",
            None,
            (
                "bytes32",
                b"",
            ),
        ),
        (
            "some_bytes",
            "bytes",
            "27",
            (
                "bytes32",
                b"X\xa2\x80\xf7OW\xbf\x05\x1c@\xf0`\x13\x9d\xc7G\xe0\x15\xbeR\xf6\x8cW\xe2\xc4\xab.K\xd4\x14oC",  # noqa: E501
            ),
        ),
        (
            "some_bytes",
            "bytes",
            27,
            (
                "bytes32",
                b"$\xd6\xd74\x14_\x07\x1a\xa6\xa2v?\xdd\xcaX\x10\xbd\x12#l->X\x9d*z\xdf\\\xa6\x9c\xc9\xc6",  # noqa: E501
            ),
        ),
        (
            "some_bytes",
            "bytes8",
            -27,
            (
                "bytes8",
                b"\x00",
            ),
        ),
        (
            "some_bytes",
            "bytes",
            -27,
            (
                "bytes32",
                b"\xbc6x\x9ez\x1e(\x146FB)\x82\x8f\x81}f\x12\xf7\xb4w\xd6e\x91\xff\x96\xa9\xe0d\xbc\xc9\x8a",  # noqa: E501
            ),
        ),
        (
            "some_bytes",
            "bytes",
            "0xdeadbeef",
            (
                "bytes32",
                b"\xd4\xfdN\x18\x912'06D\x9f\xc9\xe1\x11\x98\xc79\x16\x1bL\x01\x16\xa9\xa2\xdc\xcd\xfa\x1cI \x06\xf1",  # noqa: E501
            ),
        ),
        (
            "some_bytes",
            "bytes",
            b"\xd4\xfdN\x18\x912'06D\x9f\xc9\xe1\x11",
            (
                "bytes32",
                b"p\xa6\xe6 \xc0:\xff\xea9\x0f\xad\x1a\xab\xf8\xf7\xdc\x1b\xd1\x82\x1c\xe4\xd5\xa1,m3+\x15X\x98M\xaf",  # noqa: E501
            ),
        ),
        (
            "name",
            "string",
            27,
            (
                "bytes32",
                b"$\xd6\xd74\x14_\x07\x1a\xa6\xa2v?\xdd\xcaX\x10\xbd\x12#l->X\x9d*z\xdf\\\xa6\x9c\xc9\xc6",  # noqa: E501
            ),
        ),
        (
            "name",
            "string",
            "Bob",
            (
                "bytes32",
                b"(\xca\xc3\x18\xa8l\x8a\nj\x91V\xc2\xdb\xa2\xc8\xc266w\xba\x05\x14\xefae\x92\xd8\x15W\xe6y\xb6",  # noqa: E501
            ),
        ),
        (
            "names",
            "string[]",
            ["Bob", "Jim"],
            (
                "bytes32",
                b"@u,V\x9a<\x1a0\x1d\xf0b\xda\xc4\x98\x10/\xfb\xcf\x06\xba\x96pDC/\n\xb5\xe9\x9dB\t!",  # noqa: E501
            ),
        ),
        (
            "names",
            "string[][]",
            [["Bob", "Jim"], ["Amy", "Sal"]],
            (
                "bytes32",
                b"\xcd\x13is\x93\xf4\xf8^\xa7\x9bi\xec\x8cE\xfc\xfe\xab\xcc9p\x0c\xa8\xecR$\xbc]\xc6[\x10-1",  # noqa: E501
            ),
        ),
        (
            "a_bool",
            "bool",
            False,
            (
                "bool",
                False,
            ),
        ),
        (
            "a_bool",
            "bool",
            True,
            (
                "bool",
                True,
            ),
        ),
        (
            "a_bool",
            "bool",
            0,
            (
                "bool",
                False,
            ),
        ),
        (
            "a_bool",
            "bool",
            1,
            (
                "bool",
                True,
            ),
        ),
        (
            "a_bool",
            "bool",
            32768,
            (
                "bool",
                True,
            ),
        ),
        (
            "a_bool",
            "bool",
            -1,
            (
                "bool",
                True,
            ),
        ),
        (
            "a_bool",
            "bool",
            b"\x01",
            (
                "bool",
                True,
            ),
        ),
        (
            "a_bool",
            "bool",
            b"\x00",
            (
                "bool",
                True,
            ),
        ),
        (
            "a_bool",
            "bool",
            "true",
            (
                "bool",
                True,
            ),
        ),
        (
            "a_bool",
            "bool",
            "false",
            (
                "bool",
                True,
            ),
        ),
        (
            "a_bool",
            "bool",
            "1",
            (
                "bool",
                True,
            ),
        ),
        (
            "a_bool",
            "bool",
            "0",
            (
                "bool",
                True,
            ),
        ),
        (
            "an_address",
            "address",
            "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
            (
                "address",
                "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
            ),
        ),
        (
            "an_int16",
            "int16",
            -25,
            (
                "int16",
                -25,
            ),
        ),
        (
            "a_uint256",
            "uint256",
            1157920892373161954235709850086879078532699846656,
            (
                "uint256",
                1157920892373161954235709850086879078532699846656,
            ),
        ),
        (
            "empty_string",
            "string",
            "",
            (
                "bytes32",
                b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';{\xfa\xd8\x04]\x85\xa4p",  # noqa: E501
            ),
        ),
        (
            "expected_sample",
            "Sample",
            {
                "name": "Cow",
            },
            (
                "bytes32",
                b"\x8flivK`W\xf4my$tFGG\x9f\x99k\xbe\xb9\x8e\xcfh]\xcb\x8b\x8ak\xb9\xe5\x17\xdd",  # noqa: E501
            ),
        ),
        (
            "expected_samples",
            "Samples",
            {
                "samples": [
                    {
                        "name": "Abe",
                    },
                    {
                        "name": "Bob",
                    },
                    {
                        "name": "Cow",
                    },
                ],
            },
            (
                "bytes32",
                b"\x94}\xef\x19'\xda\xc4\x1f\x85\xfb\x87ya\x8e\x1f\x87\x96L(\x80\\\xeb\t\xe6Gv8\x0c\x19k\\\t",  # noqa: E501
            ),
        ),
        (
            "some_bytes",
            "int256",
            "0xdeadbeef",
            (
                "int256",
                3735928559,
            ),
        ),
        (
            "some_string",
            "int16",
            "412",
            (
                "int16",
                412,
            ),
        ),
        (
            "some_int",
            "int",
            3735928559,
            (
                "int",
                3735928559,
            ),
        ),
        (
            "some_uint",
            "uint",
            412,
            (
                "uint",
                412,
            ),
        ),
        (
            "a_string_array",
            "string[]",
            [],
            (
                "bytes32",
                b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';{\xfa\xd8\x04]\x85\xa4p",  # noqa: E501
            ),
        ),
        (
            "custom_type",
            "Samples",
            {"name": "bob", "samples": []},
            (
                "bytes32",
                b"(^\x89e\x13 \xa6\x16\xdf\r3\x1bOC\ngW\xb6\\\xb7\x1fn\x10\x8eh\xdc\x06\xa5\xd5\\\xc2\xb3",  # noqa: E501
            ),
        ),
    ),
    ids=[
        "None value for custom type",
        "None value for string type",
        "None value for bytes type",
        "string that is int-parseable value for bytes type",
        "int value for bytes type",
        "negative int value for bytes8 type coerces to 0 bytes8 value",
        "negative int value for bytes type coerces keccak(0) output",
        "hexstr value for bytes type",
        "bytes value for bytes type",
        "int value for string type",
        "string value for string type",
        "string[] value for string[] type",
        "string[][] value for string[][] type",
        "bool value False for bool type returns False",
        "bool value True for bool type returns True",
        "int value 0 for bool type returns False",
        "int value 1 for bool type returns True",
        "int value 37268 for bool type returns True",
        "int value -1 for bool type returns True",
        "bytes value b'\x01' for bool type returns True",
        "bytes value b'\x00' for bool type returns True",
        "string 'true' value for bool type returns True",
        "string 'false' value for bool type returns True",
        "string '1' value for bool type returns True",
        "string '0' value for bool type returns True",
        "address value for address type",
        "int16 value for int16 type",
        "uint256 value for uint256 type",
        "empty string value for string type",
        "expected value for custom type",
        "expected value for custom type array",
        "hexstr value for int256 type",
        "str value for int16 type",
        "int value for int type",
        "int value for uint type",
        "empty array value for string[] type",
        "empty array value for custom[] type",
    ],
)
def test_encode_field_pass(name, type_, value, expected):
    types = {
        "Sample": [
            {"name": "name", "type": "string"},
        ],
        "Samples": [
            {"name": "samples", "type": "Sample[]"},
        ],
    }
    assert encode_field(types, name, type_, value) == expected


@pytest.mark.parametrize(
    "name, type_, value, expected",
    (
        (
            "atomic_type",
            "address",
            None,
            {
                "expected_exception": ValueError,
                "match": "Missing value for field `atomic_type` of type `address`",
            },
        ),
        (
            "non_int_string",
            "uint256",
            "i am not an int",
            {
                "expected_exception": ValueError,
                "match": re.escape(
                    "invalid literal for int() with base 10: 'i am not an int'"
                ),
            },
        ),
        (
            "non_hex_string",
            "uint256",
            "0xi am not an int",
            {
                "expected_exception": ValueError,
                "match": re.escape(
                    "invalid literal for int() with base 10: '0xi am not an int'"
                ),
            },
        ),
        (
            "a_string",
            "string",
            [],
            {
                "expected_exception": TypeError,
                "match": re.escape(
                    "Arguments passed as hexstr or text must be of text type. Instead, value was: []"  # noqa: E501
                ),
            },
        ),
        (
            "missing_inner_array",
            "string[][]",
            ["a", "b", "c"],
            {
                "expected_exception": ValueError,
                "match": re.escape(
                    "Invalid value for field `missing_inner_array` of type `string[]`: expected array, got `a` of type `<class 'str'>`"  # noqa: E501
                ),
            },
        ),
    ),
    ids=[
        "None value for atomic type",
        "string value that is not convertible to int for int type",
        "string starting with 0x that is not convertible to int for int type",
        "empty array value for string type",
        "string[] value for string[][] type",
    ],
)
def test_encode_field_fail(name, type_, value, expected):
    types = {
        "Sample": [
            {"name": "name", "type": "string"},
        ]
    }
    with pytest.raises(**expected):
        encode_field(types, name, type_, value)


@pytest.mark.parametrize(
    "primary_type, types, expected",
    (
        (
            "Person",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            {"Person"},
        ),
        (
            "Mail",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            {"Person", "Mail"},
        ),
        (
            "Person",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friend", "type": "Person"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            {"Person"},
        ),
        (
            "Mail",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friends", "type": "Person[]"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                    {"name": "attachments", "type": "Attachment[]"},
                ],
                "Attachment": [
                    {"name": "from", "type": "string"},
                ],
            },
            {"Person", "Mail", "Attachment"},
        ),
        (
            "Ma il!",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friends", "type": "Person"},
                ],
                "Ma il!": [
                    {"name": "from", "type": "Person"},
                    {"name": "attachments", "type": "Attachment[]"},
                ],
                "Attachment": [
                    {"name": "from", "type": "string"},
                ],
            },
            {"Person", "Ma il!", "Attachment"},
        ),
    ),
    ids=[
        "type with no dependencies",
        "type with one dependency",
        "type with recursive dependency",
        "type with array dependency",
        "type with non-alphanumeric in primary type",
    ],
)
def test_find_type_dependencies_pass(primary_type, types, expected):
    assert find_type_dependencies(primary_type, types) == expected


@pytest.mark.parametrize(
    "primary_type, types, expected",
    (
        (
            27,
            {
                27: [
                    {"name": "name", "type": "string"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            {
                "expected_exception": ValueError,
                "match": "Invalid find_type_dependencies input: expected string, "
                "got `27` of type `<class 'int'>`",
            },
        ),
        (
            "Person",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "pet", "type": "Animal"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            {
                "expected_exception": ValueError,
                "match": "No definition of type `Animal`",
            },
        ),
        (
            "Mail",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friends", "type": "Person[]"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                    {"name": "attachments", "type": "Attachment[]"},
                ],
            },
            {
                "expected_exception": ValueError,
                "match": "No definition of type `Attachment`",
            },
        ),
    ),
    ids=[
        "primary_type must be a string",
        "custom type in dependencies but not defined in types",
        "custom type array in dependencies but not defined in types",
    ],
)
def test_find_type_dependencies_fail(primary_type, types, expected):
    with pytest.raises(**expected):
        find_type_dependencies(primary_type, types)


@pytest.mark.parametrize(
    "primary_type, types, encode_type_expected, hash_type_expected",
    (
        (
            "Person",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            "Person(string name)",
            "fcbb73369ebb221abfdc626fdec0be9ca48ad89ef757b9a76eb7b31ddd261338",
        ),
        (
            "Mail",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            "Mail(Person from)Person(string name)",
            "979fb836e7a3c69210ec4cfec014139732f26442be0f5c10c4222ccabfe6a025",
        ),
        (
            "Person",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friend", "type": "Person"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            "Person(string name,Person friend)",
            "945f8587e8daf26c14c625b054bfccb52dc48747984b7925c0901efca7186e14",
        ),
        (
            "Mail",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friends", "type": "Person[]"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                    {"name": "attachments", "type": "Attachment[]"},
                ],
                "Attachment": [
                    {"name": "from", "type": "string"},
                ],
            },
            "Mail(Person from,Attachment[] attachments)Attachment(string from)Person(string name,Person[] friends)",  # noqa: E501
            "0c555dd4d869769a9f68c39860155342ea7e75a215669bc44efca1b4544735e5",
        ),
        (
            "Ma il!",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friends", "type": "Person"},
                ],
                "Ma il!": [
                    {"name": "from", "type": "Person"},
                    {"name": "attachments", "type": "Attachment[]"},
                ],
                "Attachment": [
                    {"name": "from", "type": "string"},
                ],
            },
            "Ma il!(Person from,Attachment[] attachments)Attachment(string from)Person(string name,Person friends)",  # noqa: E501
            "099b05cabe783f37d88d94e349693aca4fc9ae40c51e5d2640ff040768e5b2ad",
        ),
    ),
    ids=[
        "type with no dependencies",
        "type with one dependency",
        "type with recursive dependency",
        "type with array dependency",
        "type with non-alphanumeric in primary type",
    ],
)
def test_encode_type_pass_and_hash_type(
    primary_type, types, encode_type_expected, hash_type_expected
):
    assert encode_type(primary_type, types) == encode_type_expected
    assert hash_type(primary_type, types).hex() == hash_type_expected


@pytest.mark.parametrize(
    "primary_type, types, expected",
    (
        (
            "Mail",
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "friends", "type": "Person[]"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                    {"name": "attachments", "type": "Attachment[]"},
                ],
            },
            {
                "expected_exception": ValueError,
                "match": "No definition of type `Attachment`",
            },
        ),
    ),
    ids=[
        "custom type array in dependencies but not defined in types",
    ],
)
def test_encode_type_fail(primary_type, types, expected):
    with pytest.raises(**expected):
        encode_type(primary_type, types)


@pytest.mark.parametrize(
    "domain_data, expected",
    (
        (
            {
                "name": "Ether Mail",
                "version": "1",
                "chainId": 1,
                "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
            },
            "f2cee375fa42b42143804025fc449deafd50cc031ca257e0b194a650a912090f",
        ),
        (
            {
                "name": "Ether Mail",
                "version": "1",
                "chainId": "1",
                "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
            },
            "f2cee375fa42b42143804025fc449deafd50cc031ca257e0b194a650a912090f",
        ),
        (
            {
                "name": "Ether Mail",
                "version": 1,
                "chainId": 1,
                "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
            },
            "902f609607aa38e1c768f260a84a1be97f3a9d65726d3e842fa5e36c6da393cb",
        ),
        (
            {
                "name": "Ether Mail",
                "version": "1",
                "chainId": 1,
                "verifyingContract": "0xcccccccccccccccccccccccccccccccccccccccc",
            },
            "f2cee375fa42b42143804025fc449deafd50cc031ca257e0b194a650a912090f",
        ),
        (
            {
                "name": "Ether Mail",
                "version": "1",
                "chainId": 1,
                "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
                "salt": "0xa9f4c8b7e576dc96308c361b46d32c04a00a0e5c2b0962d9f42be6891a95d139",  # noqa: E501
            },
            "53d039704f24ce448de9dc98c5952dd85b7e7c22446a0b1cb47b43b901d00972",
        ),
        (
            {},
            "6192106f129ce05c9075d319c1fa6ea9b3ae37cbd0c1ef92e2be7137bb07baa1",
        ),
    ),
    ids=[
        "EIP712 example domain data",
        "chainId as string",
        "version as int",
        "verifying contract not checksummed",
        "salt present but empty bytes",
        "empty domain data",
    ],
)
def test_hash_domain_pass(domain_data, expected):
    assert hash_domain(domain_data).hex() == expected


@pytest.mark.parametrize(
    "domain_data, expected",
    (
        (
            {
                "name": "Ether Mail",
                "classification": "1",
                "chainId": 1,
                "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
            },
            {
                "expected_exception": ValueError,
                "match": "Invalid domain key: `classification`",
            },
        ),
        (
            {
                "name": "Ether Mail",
                "version": "1",
                "chainId": 1,
                "verifyingContract": "0xCcCCccccCCCC",
            },
            {
                "expected_exception": EncodingTypeError,
                "match": "Value `'0xCcCCccccCCCC'` of type <class 'str'> cannot be encoded by AddressEncoder",  # noqa: E501
            },
        ),
    ),
    ids=[
        "key in domain_data not in EIP712 domain fields",
        "invalid address for verifyingContract",
    ],
)
def test_hash_domain_fail(domain_data, expected):
    with pytest.raises(**expected):
        hash_domain(domain_data)


@pytest.mark.parametrize(
    "type_,message, types, encode_data_expected, hash_struct_expected",
    (
        (
            "Mail",
            {
                "from": {
                    "name": "Cow",
                },
            },
            {
                "Person": [
                    {"name": "name", "type": "string"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                ],
            },
            "979fb836e7a3c69210ec4cfec014139732f26442be0f5c10c4222ccabfe6a025bfb54ec5e6bf391ea339a110356cb0fd003296b0dabe4b0b2e51d0e50c815c8a",  # noqa: E501
            "dfa5fd27fea278587b6c6a56d8e6cf2853b6698a4244afc1f5f526f04b2b70b3",
        ),
        (
            "People",
            {
                "who": [
                    {
                        "name": "Cow",
                    },
                    {
                        "name": "Dan",
                    },
                    {
                        "name": "Eve",
                    },
                ],
            },
            {
                "Person": [
                    {"name": "name", "type": "string"},
                ],
                "People": [
                    {"name": "who", "type": "Person[]"},
                ],
            },
            "cd6dded05d140e80a24f1b3f8ec7635be0235a2bb1ac5acc4f6a8c0e81959c1fdd333db219c66d1b097970cebde115462698f9bfe6ea7497898a67c7da923d33",  # noqa: E501
            "978fbd13a22cb2ced753b88943583080d6e2fa20d9f5818181dd85ee26438745",
        ),
        (
            "Things",
            {
                "what": [
                    [
                        {
                            "name": "Cow",
                        },
                    ],
                    [
                        {
                            "name": "Dan",
                        },
                    ],
                    [
                        {
                            "name": "Eve",
                        },
                    ],
                ],
            },
            {
                "Stuff": [
                    {"name": "name", "type": "string"},
                ],
                "Things": [
                    {"name": "what", "type": "Stuff[][]"},
                ],
            },
            "888cc41f0207f14b9f9e8dbaeb2cedc28dcd62ea4bd2d68c81ebb1416c846e63676cc832675b2d6ab9eb44e83f08c0df303bf2771036d926ed6fb385b6ee6de0",  # noqa: E501
            "b475420217c60fe1a7ad38c925c80f5d2c58e0fbb980684e4722f810ba9235d8",
        ),
        (
            "Things",
            {
                "what": [
                    [
                        {
                            "name": "Cow",
                        },
                    ],
                    [
                        {
                            "name": "Dan",
                        },
                    ],
                    [
                        {
                            "name": "Eve",
                        },
                    ],
                ],
            },
            {
                "Stuff": [
                    {"name": "name", "type": "string"},
                ],
                "Things": [
                    {"name": "what", "type": "Stuff[3][1]"},
                ],
            },
            "0e1a553943c15cc6227e8a3d2c788d3e74b25f85f30be99bfc6228e404e79fb9676cc832675b2d6ab9eb44e83f08c0df303bf2771036d926ed6fb385b6ee6de0",  # noqa: E501
            "cdbacf00da86992e9443d46aa0206e27d670a6140155f8c505a68ba733c7e639",
        ),
        (
            "Things",
            {
                "what": [
                    [
                        {
                            "name": "Cow",
                        },
                    ],
                    [
                        {
                            "name": "Dan",
                        },
                    ],
                    [
                        {
                            "name": "Eve",
                        },
                    ],
                ],
            },
            {
                "Stuff": [
                    {"name": "name", "type": "string"},
                ],
                "Things": [
                    {"name": "what", "type": "Stuff[8][5]"},
                ],
            },
            "5e38f404f170c80d6c941310101dfad8882e50828efc75c104c7675960b4cc94676cc832675b2d6ab9eb44e83f08c0df303bf2771036d926ed6fb385b6ee6de0",  # noqa: E501
            "ed3eb2f09fad610e8805f43a34704858e1ad7f3cd12b61e712b402be370b9001",
        ),
        (
            "Things",
            {
                "what": b"1234567890abcdef",
            },
            {
                "Things": [
                    {"name": "what", "type": "bytes16"},
                ],
            },
            "b093a949c507c3814cc23da9eebb03679633e22cec8c8633c878f43b025201333132333435363738393061626364656600000000000000000000000000000000",  # noqa: E501
            "6825950a843718a846bf289599316a041180fd20d942ae0ca6106396ff797655",
        ),
    ),
    ids=[
        "primary type with one custom type dependency",
        "primary type with one custom type array dependency",
        "primary type with one custom type nested array dependency",
        "encodes when data matches defined array size",
        "encodes when data does not match defined array size",
        "bytes8 value for bytes16 type",
    ],
)
def test_encode_data_pass_and_hash_struct_and_hash_eip712_message(
    type_,
    message,
    types,
    encode_data_expected,
    hash_struct_expected,
):
    assert encode_data(type_, types, message).hex() == encode_data_expected
    assert hash_struct(type_, types, message).hex() == hash_struct_expected
    assert hash_eip712_message(types, message).hex() == hash_struct_expected


@pytest.mark.parametrize(
    "type_,message, types, expected_error",
    (
        (
            "Things",
            {
                "what": -327,
            },
            {
                "Things": [
                    {"name": "what", "type": "uint"},
                ],
            },
            {
                "expected_exception": ValueOutOfBounds,
                "match": re.escape(
                    "Value `-327` of type <class 'int'> cannot be encoded by UnsignedIntegerEncoder: Cannot be encoded in 256 bits. Must be bounded between [0, 115792089237316195423570985008687907853269984665640564039457584007913129639935]."  # noqa: E501
                ),
            },
        ),
        (
            "Things",
            {
                "what": b"1234567890abcdef1234567890abcdef",
            },
            {
                "Things": [
                    {"name": "what", "type": "bytes8"},
                ],
            },
            {
                "expected_exception": ValueOutOfBounds,
                "match": "Value `b'1234567890abcdef1234567890abcdef'` of type <class "
                "'bytes'> cannot be encoded by BytesEncoder: exceeds total "
                "byte size for bytes8 encoding",
            },
        ),
        (
            "Things",
            {
                "what": 4294967295,
            },
            {
                "Things": [
                    {"name": "what", "type": "int16"},
                ],
            },
            {
                "expected_exception": ValueOutOfBounds,
                "match": re.escape(
                    "Value `4294967295` of type <class 'int'> cannot be encoded by SignedIntegerEncoder: Cannot be encoded in 16 bits. Must be bounded between [-32768, 32767]."  # noqa: E501
                ),
            },
        ),
        (
            "Things",
            {
                "what": "0xdeadbeef",
            },
            {
                "Things": [
                    {"name": "what", "type": "address"},
                ],
            },
            {
                "expected_exception": EncodingTypeError,
                "match": re.escape(
                    "Value `'0xdeadbeef'` of type <class 'str'> cannot be encoded by AddressEncoder",  # noqa: E501
                ),
            },
        ),
        (
            "Things",
            {
                "what": "deadbeef",
            },
            {
                "Things": [
                    {"name": "what", "type": "int256"},
                ],
            },
            {
                "expected_exception": ValueError,
                "match": re.escape(
                    "invalid literal for int() with base 10: 'deadbeef",
                ),
            },
        ),
    ),
    ids=[
        "negative value -327 for uint type",
        "bytes16 value for bytes8 type",
        "int value too large for int16 type",
        "hexstring too short to be an address",
        "hexstring must have 0x prefix be parsed as int",
    ],
)
def test_encode_data_fail(type_, message, types, expected_error):
    with pytest.raises(**expected_error):
        encode_data(type_, types, message)



from typing import (
    Any,
    Dict,
    Sequence,
)

from toolz import (
    assoc,
    dissoc,
)

from eth_account._utils.validation import (
    is_rlp_structured_access_list,
    is_rpc_structured_access_list,
)


def normalize_transaction_dict(txn_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Normalizes a transaction dictionary.
    """
    # convert all lists to tuples recursively
    for key, value in txn_dict.items():
        if isinstance(value, (list, tuple)):
            txn_dict[key] = tuple(
                normalize_transaction_dict(item) if isinstance(item, dict) else item
                for item in value
            )

        elif isinstance(value, dict):
            txn_dict[key] = normalize_transaction_dict(value)

    return txn_dict


def set_transaction_type_if_needed(transaction_dict: Dict[str, Any]) -> Dict[str, Any]:
    if "type" not in transaction_dict:
        if all(
            type_1_arg in transaction_dict for type_1_arg in ("gasPrice", "accessList")
        ):
            # access list txn - type 1
            transaction_dict = assoc(transaction_dict, "type", "0x1")
        elif any(
            type_2_arg in transaction_dict
            for type_2_arg in ("maxFeePerGas", "maxPriorityFeePerGas")
        ):
            if any(
                type_3_arg in transaction_dict
                for type_3_arg in ("maxFeePerBlobGas", "blobVersionedHashes")
            ):
                # blob txn - type 3
                transaction_dict = assoc(transaction_dict, "type", "0x3")
            else:
                # dynamic fee txn - type 2
                transaction_dict = assoc(transaction_dict, "type", "0x2")
    return transaction_dict


# JSON-RPC to rlp transaction structure
def transaction_rpc_to_rlp_structure(dictionary: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert a JSON-RPC-structured transaction to an rlp-structured transaction.
    """
    access_list = dictionary.get("accessList")
    if access_list:
        dictionary = dissoc(dictionary, "accessList")
        rlp_structured_access_list = _access_list_rpc_to_rlp_structure(access_list)
        dictionary = assoc(dictionary, "accessList", rlp_structured_access_list)
    return dictionary


def _access_list_rpc_to_rlp_structure(access_list: Sequence) -> Sequence:
    if not is_rpc_structured_access_list(access_list):
        raise ValueError(
            "provided object not formatted as JSON-RPC-structured access list"
        )

    # flatten each dict into a tuple of its values
    return tuple(
        (
            d["address"],  # value of address
            tuple(_ for _ in d["storageKeys"]),  # tuple of storage key values
        )
        for d in access_list
    )


# rlp to JSON-RPC transaction structure
def transaction_rlp_to_rpc_structure(dictionary: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert an rlp-structured transaction to a JSON-RPC-structured transaction.
    """
    access_list = dictionary.get("accessList")
    if access_list:
        dictionary = dissoc(dictionary, "accessList")
        rpc_structured_access_list = _access_list_rlp_to_rpc_structure(access_list)
        dictionary = assoc(dictionary, "accessList", rpc_structured_access_list)

    return dictionary


def _access_list_rlp_to_rpc_structure(access_list: Sequence) -> Sequence:
    if not is_rlp_structured_access_list(access_list):
        raise ValueError("provided object not formatted as rlp-structured access list")

    # build a dictionary with appropriate keys for each tuple
    return tuple({"address": t[0], "storageKeys": t[1]} for t in access_list)



from typing import (
    NamedTuple,
)
import warnings

from hexbytes import (
    HexBytes,
)


def __getitem__(self, index):
    try:
        return tuple.__getitem__(self, index)
    except TypeError:
        return getattr(self, index)


class SignedTransaction(
    NamedTuple(
        "SignedTransaction",
        [
            ("rawTransaction", HexBytes),
            ("raw_transaction", HexBytes),
            ("hash", HexBytes),
            ("r", int),
            ("s", int),
            ("v", int),
        ],
    )
):
    rawTransaction: HexBytes
    raw_transaction: HexBytes
    hash: HexBytes
    r: int
    s: int
    v: int

    def __getattribute__(cls, name):
        if name == "rawTransaction":
            warnings.warn(
                "The attribute rawTransaction on SignedTransaction is deprecated "
                "in favor of raw_transaction",
                DeprecationWarning,
                stacklevel=2,
            )
        return super().__getattribute__(name)

    def __getitem__(self, index):
        return __getitem__(self, index)


class SignedMessage(
    NamedTuple(
        "SignedMessage",
        [
            ("messageHash", HexBytes),
            ("message_hash", HexBytes),
            ("r", int),
            ("s", int),
            ("v", int),
            ("signature", HexBytes),
        ],
    )
):
    messageHash: HexBytes
    message_hash: HexBytes
    r: int
    s: int
    v: int
    signature: HexBytes

    def __getattribute__(cls, name):
        if name == "messageHash":
            warnings.warn(
                "The attribute messageHash on SignedMessage is deprecated "
                "in favor of message_hash",
                DeprecationWarning,
                stacklevel=2,
            )
        return super().__getattribute__(name)

    def __getitem__(self, index):
        return __getitem__(self, index)



from eth_utils import (
    ValidationError,
)

from .deterministic import (
    HDPath,
)
from .mnemonic import (
    Mnemonic,
)

ETHEREUM_DEFAULT_PATH = "m/44'/60'/0'/0/0"


def generate_mnemonic(num_words: int, lang: str) -> str:
    return Mnemonic(lang).generate(num_words)


def seed_from_mnemonic(words: str, passphrase: str) -> bytes:
    lang = Mnemonic.detect_language(words)
    expanded_words = Mnemonic(lang).expand(words)
    if not Mnemonic(lang).is_mnemonic_valid(expanded_words):
        raise ValidationError(
            f"Provided words: '{expanded_words}', are not a "
            "valid BIP39 mnemonic phrase!"
        )
    return Mnemonic.to_seed(expanded_words, passphrase)


def key_from_seed(seed: bytes, account_path: str) -> bytes:
    return HDPath(account_path).derive(seed)



from eth_utils.exceptions import (
    ValidationError,
)
import pytest

from eth_account.messages import (
    _hash_eip191_message,
    encode_structured_data,
)

"""
`encode_structured_data` is eth-account's take on EIP-712 typed structured data hashing.

signTypedData is MetaMask's interpretation:
https://github.com/MetaMask/eth-sig-util/tree/main

Their signTypedData_v4 has become a de facto standard.

There is no clarification in the EIP about how to handle missing or extra data.
signTypedData_v4 has taken a more permissive approach than eth-account has.

The tests below will clarify what encode_structured_data allows.
The table and tests borrow heavily from `eth-sig-util`'s:
https://github.com/MetaMask/eth-sig-util/blob/main/src/sign-typed-data.test.ts


| Input type                                   | encode_structured_data |
| -------------------------------------------- | ---------------------- |
| Example data from EIP-712                    | Y                      |
| Example data from EIP-712 w/ array           | Y                      |
| Custom type                                  | Y                      |
| Recursive custom type                        | Y                      |
| Unused custom type                           | Y*                     |
| Custom type w/ extra properties in `message` | Y*                     |
| Custom type w/ extra properties in `types`   | N                      |
| Atomic type w/ `None` input                  | N                      |
| Atomic type missing                          | N                      |
| Dynamic type w/ `None` input                 | N                      |
| Dynamic type missing                         | N                      |
| Custom type w/ `None` input                  | N                      |
| Custom type missing                          | N                      |
| Unrecognized primary type                    | N                      |

* encode_structured_data ignores fields in `message` and `domain` that are not in
`types`, and same for custom types defined in `types` that are not present in `message`.
A message containing such fields will hash, but the output will be the same as with
those fields removed.
"""


@pytest.mark.parametrize(
    "_input_type, expected_esd_hash, message_to_encode",
    (
        (
            "Example data from EIP-712",
            "be609aee343fb3c4b28e1df9e632fca64fcfaede20f02e86244efddf30957bd2",
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                        {"name": "version", "type": "string"},
                        {"name": "chainId", "type": "uint256"},
                        {"name": "verifyingContract", "type": "address"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "wallet", "type": "address"},
                    ],
                    "Mail": [
                        {"name": "from", "type": "Person"},
                        {"name": "to", "type": "Person"},
                        {"name": "contents", "type": "string"},
                    ],
                },
                "primaryType": "Mail",
                "domain": {
                    "name": "Ether Mail",
                    "version": "1",
                    "chainId": 1,
                    "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
                },
                "message": {
                    "from": {
                        "name": "Cow",
                        "wallet": "0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826",
                    },
                    "to": {
                        "name": "Bob",
                        "wallet": "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB",
                    },
                    "contents": "Hello, Bob!",
                },
            },
        ),
        (
            "Example data from EIP-712 with array",
            "1780e7e042fa9ec126ccb68cd707d61580d00601b3eff8a5ec05116b46007fdb",
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                        {"name": "version", "type": "string"},
                        {"name": "chainId", "type": "uint256"},
                        {"name": "verifyingContract", "type": "address"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "wallet", "type": "address"},
                    ],
                    "Mail": [
                        {"name": "from", "type": "Person"},
                        {"name": "to", "type": "Person"},
                        {"name": "cc", "type": "Person[]"},
                        {"name": "contents", "type": "string"},
                    ],
                },
                "primaryType": "Mail",
                "domain": {
                    "name": "Ether Mail",
                    "version": "1",
                    "chainId": 1,
                    "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
                },
                "message": {
                    "from": {
                        "name": "Cow",
                        "wallet": "0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826",
                    },
                    "to": {
                        "name": "Bob",
                        "wallet": "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB",
                    },
                    "cc": [
                        {
                            "name": "Alice",
                            "wallet": "0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
                        },
                        {
                            "name": "Dot",
                            "wallet": "0xdddddddddddddddddddddddddddddddddddddddd",
                        },
                    ],
                    "contents": "Hello, Bob!",
                },
            },
        ),
        (
            "Custom type",
            "2f1c830ad734e08c06419c229b6bb26e0fb4134b5c04a4dea2163171b79683eb",
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                },
            },
        ),
        (
            "Recursive custom type",
            "36c7cd01046282632deb49c4095ec5c504502369acfed60c1120b7d7f32740a9",
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "friends", "type": "Person[]"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                    "friends": [
                        {
                            "name": "Charlie",
                            "friends": [],
                        }
                    ],
                },
            },
        ),
        (
            "Unused custom type",
            "2f1c830ad734e08c06419c229b6bb26e0fb4134b5c04a4dea2163171b79683eb",
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                    ],
                    "Pet": [
                        {"name": "animal", "type": "string"},
                        {"name": "age", "type": "uint256"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                },
            },
        ),
        (
            "Custom type with extra properties in message",
            "2f1c830ad734e08c06419c229b6bb26e0fb4134b5c04a4dea2163171b79683eb",
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                    "age": 33,
                },
            },
        ),
    ),
)
def test_encode_structured_data_pass(_input_type, expected_esd_hash, message_to_encode):
    assert (
        _hash_eip191_message(encode_structured_data(message_to_encode)).hex()
        == expected_esd_hash
    )


@pytest.mark.parametrize(
    "_input_type, expected_esd_err, message_to_encode",
    (
        (
            "Custom type with extra properties in types",
            {
                "expected_exception": KeyError,
                "match": "age",
            },
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "age", "type": "uint256"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                },
            },
        ),
        (
            "Atomic type with `None` input",
            {
                "expected_exception": ValueError,
                "match": "Missing value for field address of type address",
            },
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "address", "type": "address"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                    "address": None,
                },
            },
        ),
        (
            "Atomic type missing",
            {
                "expected_exception": KeyError,
                "match": "address",
            },
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "address", "type": "address"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                },
            },
        ),
        (
            "Dynamic type with `None` input",
            {
                "expected_exception": ValueError,
                "match": "Missing value for field motto of type string",
            },
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "motto", "type": "string"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                    "motto": None,
                },
            },
        ),
        (
            "Dynamic type missing",
            {
                "expected_exception": KeyError,
                "match": "motto",
            },
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "motto", "type": "string"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                },
            },
        ),
        (
            "Custom type with `None` input",
            {
                "expected_exception": ValueError,
                "match": "Missing value for field friend of type Person",
            },
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "friend", "type": "Person"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                    "friend": None,
                },
            },
        ),
        (
            "Custom type missing",
            {
                "expected_exception": KeyError,
                "match": "friend",
            },
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "friend", "type": "Person"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                },
            },
        ),
        (
            "Unrecognized primary type",
            {
                "expected_exception": ValidationError,
                "match": "The Primary Type `Person` is not present in the `types` attribute",  # noqa: E501
            },
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                    ],
                    "Human": [
                        {"name": "name", "type": "string"},
                    ],
                },
                "primaryType": "Person",
                "domain": {
                    "name": "Name",
                },
                "message": {
                    "name": "Bob",
                },
            },
        ),
    ),
)
def test_encode_structured_data_fail(_input_type, expected_esd_err, message_to_encode):
    with pytest.raises(**expected_esd_err):
        encode_structured_data(message_to_encode)



from .access_list_transaction import (
    AccessListTransaction,
)
from .blob_transactions.blob_transaction import (
    BlobTransaction,
)
from .dynamic_fee_transaction import (
    DynamicFeeTransaction,
)
from .typed_transaction import (
    TypedTransaction,
)



import itertools
from typing import (
    Dict,
)

from eth_rlp import (
    HashableRLP,
)
from eth_utils.curried import (
    apply_formatters_to_dict,
)
from eth_utils.toolz import (
    curry,
    dissoc,
    merge,
    partial,
    pipe,
)
import rlp
from rlp.sedes import (
    Binary,
    big_endian_int,
    binary,
)

from .transaction_utils import (
    set_transaction_type_if_needed,
)
from .typed_transactions import (
    TypedTransaction,
)
from .validation import (
    LEGACY_TRANSACTION_FORMATTERS,
    LEGACY_TRANSACTION_VALID_VALUES,
)


def serializable_unsigned_transaction_from_dict(transaction_dict, blobs=None):
    transaction_dict = set_transaction_type_if_needed(transaction_dict)
    if "type" in transaction_dict:
        # We delegate to TypedTransaction, which will carry out validation & formatting.
        return TypedTransaction.from_dict(transaction_dict, blobs=blobs)

    if blobs is not None:
        # sanity check, blobs should never get past typed transactions check above
        raise TypeError("Blob data is not supported for legacy transactions.")

    assert_valid_fields(transaction_dict)
    filled_transaction = pipe(
        transaction_dict,
        dict,
        partial(merge, TRANSACTION_DEFAULTS),
        chain_id_to_v,
        apply_formatters_to_dict(LEGACY_TRANSACTION_FORMATTERS),
    )
    if "v" in filled_transaction:
        serializer = Transaction
    else:
        serializer = UnsignedTransaction
    return serializer.from_dict(filled_transaction)


def encode_transaction(unsigned_transaction, vrs):
    (v, r, s) = vrs
    chain_naive_transaction = dissoc(unsigned_transaction.as_dict(), "v", "r", "s")
    if isinstance(unsigned_transaction, TypedTransaction):
        # Typed transaction have their own encoding format,
        # so we must delegate the encoding.
        chain_naive_transaction["v"] = v
        chain_naive_transaction["r"] = r
        chain_naive_transaction["s"] = s
        blob_data = unsigned_transaction.blob_data
        signed_typed_transaction = TypedTransaction.from_dict(
            chain_naive_transaction,
            blobs=[blob.as_bytes() for blob in blob_data.blobs] if blob_data else None,
        )
        return signed_typed_transaction.encode()
    signed_transaction = Transaction(v=v, r=r, s=s, **chain_naive_transaction)
    return rlp.encode(signed_transaction)


TRANSACTION_DEFAULTS = {
    "to": b"",
    "value": 0,
    "data": b"",
    "chainId": None,
}

ALLOWED_TRANSACTION_KEYS = {
    "nonce",
    "gasPrice",
    "gas",
    "to",
    "value",
    "data",
    # set chainId to None if you want a transaction that can be replayed across networks
    "chainId",
}

REQUIRED_TRANSACTION_KEYS = ALLOWED_TRANSACTION_KEYS.difference(
    TRANSACTION_DEFAULTS.keys()
)


def assert_valid_fields(transaction_dict):
    # check if any keys are missing
    missing_keys = REQUIRED_TRANSACTION_KEYS.difference(transaction_dict.keys())
    if missing_keys:
        raise TypeError(f"Transaction must include these fields: {repr(missing_keys)}")

    # check if any extra keys were specified
    superfluous_keys = set(transaction_dict.keys()).difference(ALLOWED_TRANSACTION_KEYS)
    if superfluous_keys:
        raise TypeError(
            "Transaction must not include unrecognized fields: "
            f"{repr(superfluous_keys)}"
        )

    # check for valid types in each field
    valid_fields: Dict[str, bool]
    valid_fields = apply_formatters_to_dict(
        LEGACY_TRANSACTION_VALID_VALUES, transaction_dict
    )
    if not all(valid_fields.values()):
        invalid = {
            key: transaction_dict[key]
            for key, valid in valid_fields.items()
            if not valid
        }
        raise TypeError(f"Transaction had invalid fields: {repr(invalid)}")


def chain_id_to_v(transaction_dict):
    # See EIP 155
    chain_id = transaction_dict.pop("chainId")
    if chain_id is None:
        return transaction_dict
    else:
        return dict(transaction_dict, v=chain_id, r=0, s=0)


@curry
def fill_transaction_defaults(transaction):
    return merge(TRANSACTION_DEFAULTS, transaction)


UNSIGNED_TRANSACTION_FIELDS = (
    ("nonce", big_endian_int),
    ("gasPrice", big_endian_int),
    ("gas", big_endian_int),
    ("to", Binary.fixed_length(20, allow_empty=True)),
    ("value", big_endian_int),
    ("data", binary),
)


class Transaction(HashableRLP):
    fields = UNSIGNED_TRANSACTION_FIELDS + (
        ("v", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    )


class UnsignedTransaction(HashableRLP):
    fields = UNSIGNED_TRANSACTION_FIELDS


ChainAwareUnsignedTransaction = Transaction


def strip_signature(txn):
    unsigned_parts = itertools.islice(txn, len(UNSIGNED_TRANSACTION_FIELDS))
    return list(unsigned_parts)


def vrs_from(transaction):
    return (getattr(transaction, part) for part in "vrs")
